import collections

import numpy as np
import scipy
from .. import Numputils as nput
from .. import Coordinerds as coordops
from .. import Iterators as itut
from .Elements import *
from .PointGroups import PointGroup
from .SymmetryIdentifier import PointGroupIdentifier


__all__ = [
    "symmetrize_structure",
    "symmetrized_coordinate_coefficients",
    "get_internal_permutation_symmetry_matrices",
    "symmetrize_internals",
    "symmetry_projected_coordinates"
]

def _symmetry_reduce(coords, op:np.ndarray, labels=None):
    perm = nput.symmetry_permutation(coords, op)
    cycles = nput.permutation_cycles(perm, return_groups=True)
    coords = np.array([
        coords[p[0]]
        for p in cycles
    ])
    if labels is not None:
        labels = [
            labels[p[0]]
            for p in cycles
        ]
        return coords, labels
    else:
        return coords

def prep_symmetry_operations(symmetry_elements: 'PointGroup|list[SymmetryElement|np.ndarray]'):
    if isinstance(symmetry_elements, PointGroup):
        symmetry_elements = symmetry_elements.elements
    return [
        e.get_transformation()
            if isinstance(e, SymmetryElement) else
        np.asanyarray(e)
        for e in symmetry_elements
    ]

# def apply_symmetry_elements(coords, symmetry_elements: 'PointGroup|list[SymmetryElement|np.ndarray]', labels=None, tol=1e-1):
#     coords = np.asanyarray(coords)
#     symmetry_elements = prep_symmetry_operations(symmetry_elements)
#
#     for e in symmetry_elements:
#         new_coords = coords @ e
#         coord_diffs = np.linalg.norm(coords[:, np.newaxis, :] - new_coords[np.newaxis, :, :], axis=-1)
#         dupe_pos = np.where(coord_diffs < tol)
#         new_labs = None
#         if len(dupe_pos) > 0 and len(dupe_pos[0]) > 0:
#             rem = np.setdiff1d(np.arange(len(new_coords)), dupe_pos[0])
#             if labels is not None:
#                 new_labs = [labels[l] for l in rem]
#             new_coords = new_coords[rem,]
#         elif labels is not None:
#             new_labs = labels
#         if labels is not None:
#             labels = labels + new_labs
#         coords = np.concatenate([coords, new_coords], axis=0)
#
#     if labels is not None:
#         return coords, labels
#     else:
#         return coords

def symmetrize_structure(coords,
                         symmetry_elements: 'PointGroup|list[SymmetryElement|np.ndarray]',
                         labels=None,
                         masses=None,
                         groups=None,
                         tol=1e-1,
                         mass_tol=1,
                         expand=True
                         ):
    coords = np.asanyarray(coords)
    if groups is None:
        if labels is not None:
            label_map = {}
            g_vec = []
            for l in labels:
                label_map[l] = label_map.get(l, len(label_map))
                g_vec.append(l)
            (_, groups), _ = nput.group_by(np.arange(len(g_vec)), g_vec)
        elif masses is not None:
            (_, groups), _ = nput.group_by(np.arange(len(masses)), np.round(masses/mass_tol))
        else:
            groups = [np.arange(len(coords))]
    symmetry_elements = prep_symmetry_operations(symmetry_elements)

    all_coords = []
    all_labels = [] if labels is not None else None
    for g in groups:
        subcoords = coords[g,]
        if labels is not None:
            sublabels = [labels[gg] for gg in g]
        else:
            sublabels = None
        for e in symmetry_elements:
            subcoords = _symmetry_reduce(subcoords, e, labels=sublabels)
            if sublabels is not None:
                subcoords, sublabels = subcoords
        if expand:
            subcoords = nput.apply_symmetries(subcoords,
                                              prep_symmetry_operations(symmetry_elements),
                                              labels=sublabels,
                                              tol=tol)
            if sublabels is not None:
                subcoords, sublabels = subcoords
        if all_labels is not None:
            all_labels.append(sublabels)
        all_coords.append(subcoords)

    all_coords = np.concatenate(all_coords)
    all_labels = sum(all_labels, [])
    if labels is not None:
        return all_coords, all_labels
    else:
        return all_coords


def symmetrized_coordinate_coefficients(point_group,
                                        coords,
                                        masses=None,
                                        permutation_basis=None,
                                        as_characters=True,
                                        normalize=False,
                                        perms=None,
                                        ops=None,
                                        return_basis=None,
                                        merge_equivalents=None,
                                        drop_empty_modes=None,
                                        realign=True,
                                        permutation_tol=1e-2,
                                        **pg_tols
                                        ):

    if perms is None:
        if ops is None:
            if isinstance(point_group, str):
                point_group = PointGroup.from_name(point_group)
            elif isinstance(point_group, (list, tuple)) and isinstance(point_group[0], str):
                point_group = PointGroup.from_name(*point_group)

            id = PointGroupIdentifier(coords, masses, **pg_tols)
            if hasattr(point_group, "character_table"):
                if realign:
                    point_group = id.embed_point_group(point_group)
                ops = point_group.get_matrices(only_class_representatives=False)
                class_perm = np.concatenate(point_group.character_table.classes, axis=0)
            elif hasattr(point_group, 'get_full_matrices'):
                ops = point_group.get_full_matrices()
                if realign:
                    elems = [
                        SymmetryElement.from_transformation_matrix(x)
                        for x in point_group.matrices
                    ]
                    axes = id.find_point_group_alignment_axes(elems)
                    base_axes = PointGroup.get_axes_from_symmetry_elements(elems)
                    tf = axes.T @ base_axes
                    ops = tf[np.newaxis] @ ops @ tf.T[np.newaxis]
                class_perm = np.concatenate(point_group.classes, axis=0)
            else:
                ops = point_group
                class_perm = np.arange(len(point_group))

            ops = ops[class_perm,]

        # TODO: find a cleaner way to do this...
        perms = np.array([
            nput.symmetry_permutation(coords, m, tol=permutation_tol)
            for m in ops
        ])

    if permutation_basis is None:
        full_modes = np.zeros((3 * perms.shape[1],) + coords.shape + (len(ops),))

        inv = np.argsort(perms, axis=1)
        for n, (p, m) in enumerate(zip(inv, ops)):
            for i, j in enumerate(p):
                full_modes[3 * i:(3 * i + 3), j, :, n] = m
        full_basis = None
    else:
        full_modes, full_basis = permutation_basis(perms)

    equivalent_coords = None
    if merge_equivalents is None:
        merge_equivalents = as_characters
    if merge_equivalents is True:
        mask = np.moveaxis(np.abs(full_modes) > 1e-6, -1, 0)
        graph = mask[0]
        for m in mask:
            graph = np.logical_or(graph, m)  # TODO: speed this up with better masks

        _, comps = scipy.sparse.csgraph.connected_components(graph.reshape(graph.shape[0], -1))
        (_, equivalent_coords) = nput.group_by(
            np.arange(len(graph)),
            comps
        )[0]

    if as_characters:
        ct = point_group
        if hasattr(ct, "character_table"):
            ct = point_group.character_table

        full_modes = np.tensordot(ct.extend_class_representation(ct.table), full_modes, axes=[-1, -1])
        if normalize:
            full_modes = nput.vec_normalize(
                full_modes.reshape(full_modes.shape[:2] + (-1,)),
                axis=-1
            ).reshape(full_modes.shape)

        if drop_empty_modes is None:
            drop_empty_modes = normalize

        if equivalent_coords is not None:
            full_modes = full_modes.reshape(full_modes.shape[:2] + (-1,))
            rep_dims = ct.table[:, 0].astype(int)

            new_modes = []
            for character_modes, character_dim in zip(full_modes, rep_dims):
                comp_bits = [x for c in equivalent_coords for x in c[:character_dim]]
                new_modes.append(character_modes[:, comp_bits])
            full_modes = new_modes

        if drop_empty_modes:
            where_pos = lambda m:np.where(np.linalg.norm(m, axis=0) > 1e-12)[0]
            full_modes = [
                modes[:, where_pos(modes)]
                for modes in full_modes
            ]
    else:
        full_modes = np.moveaxis(full_modes, -1, 0)
        if normalize:
            full_modes = nput.vec_normalize(
                full_modes.reshape(full_modes.shape[:2] + (-1,)),
                axis=-1
            ).reshape(full_modes.shape)

        if drop_empty_modes is None:
            drop_empty_modes = normalize

        if equivalent_coords is not None:
            comp_bits = [c[0] for c in equivalent_coords]
            full_modes = full_modes.reshape((full_modes.shape[0], full_modes.shape[-1], -1))
            full_modes = full_modes[:, :, comp_bits]

        if drop_empty_modes:
            where_pos = lambda modes:np.where(np.linalg.norm(modes, axis=0) > 1e-12)[0]
            full_modes = [
                modes[:, where_pos(modes)]
                for modes in full_modes
            ]

    if return_basis is None:
        return_basis = full_basis is not None

    if return_basis:
        return full_modes, full_basis
    else:
        return full_modes

def get_internal_permutation_symmetry_matrices(internals, permutations):
    map = {
        coordops.canonicalize_internal(i): n
        for n, i in enumerate(internals)
    }
    internals = list(map.keys())
    basis = []
    p_coords = internals
    while len(p_coords) > 0:
        new_coords = []
        for n,p in enumerate(permutations):
            if len(basis) < n+1:
                subbasis = [
                    [0] * len(internals)
                    for _ in range(len(internals))
                ]
                basis.append(subbasis) # take advantage of mutability
            else:
                subbasis = basis[n]
            new_ints = coordops.permute_internals(p_coords, p, canonicalize=True)
            for old,new in zip(p_coords, new_ints):
                j = map[old]
                if new == old:
                    subbasis[j][j] = 1
                elif (
                        len(new) == 4
                        and new[0] == old[0] and new[3] == old[3]
                        and new[1] == old[2] and new[2] == old[1]
                ):
                    subbasis[j][j] = 1
                else:
                    if len(old) < 4:
                        sign = 1
                    else:
                        sign = 1

                    if (i := map.get(new)) is None:
                        if len(new) == 4:
                            for o,k in map.items():
                                if (
                                        len(o) == 4
                                        and new[0] == o[0] and o[3] == o[3]
                                        and new[1] == o[2] and new[2] == o[1]
                                ):
                                    i = k
                                    # sign *= -1
                        if i is None:
                            for sb in basis:
                                sb.append([0] * (len(internals)))
                                for b in sb: b.append(0)
                            i = len(internals)
                            map[new] = i
                            internals.append(new)
                            new_coords.append(new)
                    subbasis[j][i] = sign
        p_coords = new_coords

    #TODO: handle newly add internals, we need their permutation symmetries too

    full_basis = []
    for subbasis in basis:
        if len(subbasis) < len(internals):
            subbasis.extend(
                [[0]*len(internals)]
                * (len(internals) - len(subbasis))
            )

        full_basis.append([
            s + [0]*(len(internals) - len(s))
            for s in subbasis
        ])

    return np.moveaxis(np.array(full_basis, dtype=int), 0, -1), internals

def symmetrize_internals(point_group,
                         internals,
                         cartesians=None,
                         *,
                         masses=None,
                         as_characters=True,
                         normalize=None,
                         perms=None,
                         return_expansions=False,
                         return_base_expansion=False,
                         ops=None,
                         atom_selection=None,
                         reduce_redundant_coordinates=None,
                         **etc
                         ):
    if cartesians is not None:
        cartesians = np.asanyarray(cartesians)
        if np.issubdtype(cartesians.dtype, np.integer):
            # we'll support permutations as Cartesians even if it feels like a bad idea
            # it's just so convenient...
            atom_list = coordops.coordinate_indices(internals)
            nats = len(atom_list)
            if cartesians.shape != (nats, 3):
                if nats == 3:
                    # we'll check to see if we have permutations...
                    test = np.sort(cartesians, axis=0)
                    if np.sum(np.abs(test - np.arange(nats)[:, np.newaxis])) < 1e-8:
                        perms, cartesians = cartesians, None
            elif cartesians.shape[-1] == nats:
                perms, cartesians = cartesians, None
        # if cartesians.shape ==
    if perms is None and cartesians is None:
        raise ValueError("either Cartesians or explicit set of atom permutations required")

    if normalize is None:
        normalize = return_expansions

    if reduce_redundant_coordinates is None:
        reduce_redundant_coordinates = return_expansions and normalize

    if return_expansions and cartesians is None:
        raise ValueError("Cartesians are required to return coordinate expansions")
    if reduce_redundant_coordinates and not return_expansions:
        raise ValueError("Expansions are required to generate redundant coordinates")

    base_carts = cartesians
    base_masses = masses
    if atom_selection is not None and cartesians is not None:
        internal_reindexing = np.argsort(np.concatenate([
            atom_selection,
            np.delete(np.arange(len(cartesians)), atom_selection)
        ]))
        cartesians = cartesians[atom_selection,]
        internals = coordops.permute_internals(internals, internal_reindexing)
        if masses is not None:
            masses = np.asanyarray(masses)[atom_selection,]
    else:
        internal_reindexing = None


    symm = lambda p: get_internal_permutation_symmetry_matrices(internals, p)
    coeffs, full_basis = symmetrized_coordinate_coefficients(
        point_group,
        cartesians,
        masses=masses,
        permutation_basis=symm,
        as_characters=as_characters,
        normalize=normalize,
        perms=perms,
        ops=ops,
        **etc
    )

    if internal_reindexing is not None:
        internal_reindexing = np.argsort(internal_reindexing)
        full_basis = coordops.permute_internals(full_basis, internal_reindexing)

    res = (coeffs, full_basis)
    if return_expansions:
        if return_expansions is True: return_expansions = 1
        normalized_coefficients = normalize is not False and (as_characters or normalize)
        expansions = symmetrized_internal_coordinate_expansions(
            coeffs,
            base_carts,
            full_basis,
            order=return_expansions,
            masses=base_masses,
            normalized_coefficients=normalized_coefficients,
            return_base_expansion=return_base_expansion,
            return_inverse=True
        )

        if isinstance(reduce_redundant_coordinates, dict):
            redundant_opts = reduce_redundant_coordinates
            reduce_redundant_coordinates = True
        else:
            redundant_opts = {}

        if reduce_redundant_coordinates:
            if return_base_expansion:
                expansions, base_expansions = expansions
            else:
                base_expansions = None
            all_exps = [e[0] for e in expansions]
            all_tfs = [np.concatenate(e, axis=-1) for e in itut.transpose(all_exps)]
            all_inv = [np.concatenate(e, axis=0) for e in itut.transpose([e[1] for e in expansions])]
            # all_coeffs = np.concatenate(coeffs, axis=-1)
            red_coeffs, red_exps = coordops.RedundantCoordinateGenerator.get_redundant_transformation(
                all_tfs[1:],
                masses=base_masses,
                **collections.ChainMap(redundant_opts, dict(relocalize=True))
            )
            red_inv = nput.inverse_internal_coordinate_tensors(
                red_exps,
                base_carts,
                masses=base_masses,
                remove_translation_rotation=True
            )
            red_exps = red_coeffs, (red_exps, red_inv)
            expansions = (red_exps, expansions)
            if base_expansions is not None:
                expansions = expansions + (base_expansions,)
        elif not return_base_expansion:
            expansions = (expansions,)

        res = res + expansions

    return res

def symmetrized_internal_coordinate_expansions(coeffs, cartesians, full_basis,
                                               order=1,
                                               masses=None,
                                               return_inverse=False,
                                               normalized_coefficients=True,
                                               return_base_expansion=False,
                                               ):

    exps = nput.internal_coordinate_tensors(cartesians, full_basis,
                                                return_inverse=return_inverse,
                                                masses=masses,
                                                order=order
                                                )

    if return_inverse:
        exp, inv = exps
        coeff_inv = [
            c.T
                if normalized_coefficients else
            np.linalg.pinv(c)
            for c in coeffs
        ]
        expansions = [
            (
                [np.dot(exp[0], c)] + nput.tensor_reexpand(exp[1:], [c]),
                nput.tensor_reexpand([ci], inv)
            )
            for c, ci in zip(coeffs, coeff_inv)
        ]
    else:
        exp = exps
        expansions = [
            [np.dot(exp[0], c)] + nput.tensor_reexpand(exp[1:], [c])
            for c in coeffs
        ]

    if return_base_expansion:
        return expansions, exps
    else:
        return expansions

    # return symm_coeffs, storage[1]

def symmetry_projected_coordinates():
    # get a relocalized delocalized coordinate set projected in the space of symmetric
    # coordinates
    ...



import abc, numpy as np, scipy.signal, math
import itertools as it
from ..Combinatorics import UniquePermutations, Binomial
from ..Numputils import SparseArray, group_by

__all__ = [
    "AbstractPolynomial",
    "DensePolynomial",
    "SparsePolynomial",
    "PureMonicPolynomial",
    "TensorCoefficientPoly"
]

class AbstractPolynomial(metaclass=abc.ABCMeta):
    """
    Provides the general interface an abstract polynomial needs ot support, including
    multiplication, addition, shifting, access of coefficients, and evaluation
    """

    @property
    @abc.abstractmethod
    def scaling(self):
        raise NotImplementedError("abstract base class")

    @abc.abstractmethod
    def __mul__(self, other)->'AbstractPolynomial':
        raise NotImplementedError("abstract base class")
    @abc.abstractmethod
    def __add__(self, other)->'AbstractPolynomial':
        raise NotImplementedError("abstract base class")

    @abc.abstractmethod
    def shift(self, shift)->'AbstractPolynomial':
        raise NotImplementedError("abstract base class")

    def __rmul__(self, other)->'AbstractPolynomial':
        return self * other
    def __radd__(self, other)->'AbstractPolynomial':
        return self + other
    def __truediv__(self, other)->'AbstractPolynomial':
        return self * (1/other)
    def __neg__(self):
        return -1*self
    def __sub__(self, other):
        return self + (-other)
    def __rsub__(self, other):
        return (-self)+other


    # def __rtruediv__(self, other):
    #     return other * (1/self)


class DensePolynomial(AbstractPolynomial):
    """
    A straightforward dense n-dimensional polynomial data structure with
    multiplications and shifts
    """
    def __init__(self,
                 coeffs,
                 prefactor=None,
                 shift=None,
                 stack_dim=0
                 ):
        self._scaling = prefactor
        self._shift = shift
        self._poly_coeffs = np.asanyarray(coeffs) if not isinstance(coeffs, SparseArray) else coeffs
        self._coeff_tensors = None
        self._uc_tensors = None
        self.stack_dim = stack_dim
    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.coeffs.shape, self.scaling)
    @classmethod
    def from_tensors(cls, tensors, prefactor=None, shift=None, rescale=True):
        coeffs, stack_dim = cls.condense_tensors(tensors, rescale=rescale)
        new = cls(coeffs, prefactor=prefactor, shift=shift, stack_dim=stack_dim)
        return new

    @property
    def shape(self):
        return self._poly_coeffs.shape

    @property
    def scaling(self):
        return 1 if self._scaling is None else self._scaling
    @scaling.setter
    def scaling(self, s):
        self._scaling = s

    @property
    def coeffs(self) -> 'np.ndarray|SparseArray':
        if self._shift is not None:
            self._poly_coeffs = self.compute_shifted_coeffs(self._poly_coeffs, self._shift, stack_dim=self.stack_dim)
            self._shift = None
        if self._scaling is not None:
            self._poly_coeffs = self._poly_coeffs * self._scaling
            self.scaling = None
        return self._poly_coeffs
    @coeffs.setter
    def coeffs(self, cs):
        self._poly_coeffs = cs

    @property
    def coordinate_dim(self):
        return self._poly_coeffs.ndim - self.stack_dim

    @classmethod
    def _manage_stack_bcast(cls, c1, c2, s1, s2):
        if s1 == s2:
            if s1 == 0:
                return c1, c2, 0
            sd = s1
            final_shape = np.broadcast_shapes(c1.shape[:s1], c2.shape[:s2])
            if isinstance(c1, np.ndarray):
                c1 = np.broadcast_to(c1, final_shape + c1.shape[s1:])
            else:
                c1 = c1.broadcast_to(final_shape + c1.shape[s1:])
            if isinstance(c2, np.ndarray):
                c2 = np.broadcast_to(c2, final_shape + c2.shape[s2:])
            else:
                c2 = c2.broadcast_to(final_shape + c2.shape[s2:])
        elif s1 == 0:
            sd = s2
            if isinstance(c1, np.ndarray):
                c1 = np.broadcast_to(np.expand_dims(c1, list(range(s2))), c2.shape[:s2] + c1.shape)
            else:
                c1 = c1.expand_and_broadcast(
                    list(range(s2)),
                    c2.shape[:s2] + c1.shape
                )
        elif s2 == 0:
            sd = s1
            if isinstance(c2, np.ndarray):
                c2 = np.broadcast_to(np.expand_dims(c2, list(range(s1))), c1.shape[:s1] + c2.shape)
            else:
                c2 = c2.expand_and_broadcast(
                    list(range(s1)),
                    c1.shape[:s1] + c2.shape
                )
        else:
            raise ValueError("can't broadcast shapes {} and {}".format(
                c1.shape[:s1],
                c2.shape[:s2]
            ))

        sh1 = c1.shape
        sh2 = c2.shape
        padding = len(sh2) - len(sh1)
        if padding > 0:
            if isinstance(c1, np.ndarray):
                c1 = np.expand_dims(c1, [-x for x in range(1, padding+1)])
            else:
                c1 = c1.expand_dims([-x for x in range(1, padding+1)])
        elif padding < 0:
            padding = -padding
            if isinstance(c2, np.ndarray):
                c2 = np.expand_dims(c2, [-x for x in range(1, padding+1)])
            else:
                c2 = c2.expand_dims([-x for x in range(1, padding+1)])

        return c1, c2, sd

    @classmethod
    def _sparse_convolve_vals_inds(cls,
                                   vals1, vals2,
                                   inds1, inds2,
                                   ):
        # print(inds1)
        # print(inds2)
        vals = (vals1[:, np.newaxis] * vals2[np.newaxis, :]).flatten()
        inds = np.array([
            (i1[:, np.newaxis] + i2[np.newaxis, :]).flatten()
            for i1, i2 in zip(inds1, inds2)
        ])
        # print(inds)
        # print(vals)
        (inds, valsets) = group_by(vals, inds.T)[0]
        vals = np.array([np.sum(v) for v in valsets])
        return inds.T, vals
    @classmethod
    def _sparse_convolve_single(cls, coeffs_1, coeffs_2):
        v1, i1 = coeffs_1.block_data
        v2, i2 = coeffs_2.block_data
        inds, vals = cls._sparse_convolve_vals_inds(
            v1, v2,
            i1, i2
        )
        shp = tuple(c1 + c2 - 1 for c1,c2 in zip(coeffs_1.shape, coeffs_2.shape))
        return SparseArray.from_data(
            (vals, tuple(inds)),
            shape=shp
        )

    @classmethod
    def _sparse_convolve_stacks(cls, coeffs_1, coeffs_2, stack_dim):
        v1, i1 = coeffs_1.block_data
        v2, i2 = coeffs_2.block_data

        stacks_1, inds_1 = group_by(
            np.arange(len(v1)),
            np.array(i1[:stack_dim]).T
        )[0]
        stacks_2, inds_2 = group_by(
            np.arange(len(v2)),
            np.array(i2[:stack_dim]).T
        )[0]

        ind_blocks = [

        ]
        val_blocks = [

        ]
        prev_match = 0
        for s1, pos1 in zip(stacks_1, inds_1):
            for j in range(prev_match, len(stacks_2)):
                if np.all(s1 == stacks_2[j]):
                    prev_match = j
                    pos2 = inds_2[j]
                    if len(pos1) > 0 and len(pos2) > 0:
                        subinds, subvals = cls._sparse_convolve_vals_inds(
                            v1[pos1], v2[pos2],
                            [x[pos1] for x in i1[stack_dim:]], [y[pos2] for y in i2[stack_dim:]]
                        )
                        # now we need to stick copy s1 onto every val to get the stacking right
                        header = tuple(np.full(len(subvals), s) for s in s1)
                        ind_blocks.append(header + tuple(subinds))
                        val_blocks.append(subvals)
                    break

        vals = np.concatenate(val_blocks)
        inds = tuple(
            np.concatenate([ib[j] for ib in ind_blocks])
            for j in range(len(i1))
        )

        shp = coeffs_1.shape[:stack_dim] + tuple(c1 + c2 - 1
                    for c1, c2 in zip(coeffs_1.shape[stack_dim:], coeffs_2.shape[stack_dim:])
                    )
        return SparseArray.from_data(
            (vals, inds),
            shape=shp
        )
    def __mul__(self, other)->'DensePolynomial':
        if isinstance(other, (int, float, np.integer, np.floating)):
            if other == 1:
                return self
            elif other == 0:
                return 0
        if isinstance(other, DensePolynomial):
            # print(self.coeffs.shape, other.coeffs.shape)
            scs, ocs, sd = self._manage_stack_bcast(
                self.coeffs, other.coeffs,
                self.stack_dim, other.stack_dim
            )
            if sd > 0:
                if isinstance(scs, np.ndarray):
                    new_coeffs = np.array([
                        scipy.signal.convolve(s, o)
                        for s,o in zip(
                            scs.reshape((-1,) + scs.shape[sd:]),
                            ocs.reshape((-1,) + ocs.shape[sd:]),
                        )
                    ])
                    new_coeffs = new_coeffs.reshape(scs.shape[:sd] + new_coeffs.shape[1:])
                else:
                    new_coeffs = self._sparse_convolve_stacks(scs, ocs, sd)
            else:
                if isinstance(scs, np.ndarray):
                    new_coeffs = scipy.signal.convolve(scs, ocs)
                else:
                    new_coeffs = self._sparse_convolve_single(scs, ocs)
            return DensePolynomial(
                new_coeffs,
                shift=None,
                prefactor=self.scaling * other.scaling,
                stack_dim=sd
            )
        else:
            return DensePolynomial(
                self._poly_coeffs,
                shift=self._shift,
                prefactor=self.scaling * other
            )

    def _force_consistent_addition_shapes(self, other):
        if self.shape[:self.stack_dim] != other.shape[:other.stack_dim]:
            fcs = self.coeffs #type: np.ndarray | SparseArray
            ocs = other.coeffs #type: np.ndarray | SparseArray
            bcast_shape = (np.zeros(self.shape[:self.stack_dim]) * np.zeros(other.shape[:other.stack_dim])).shape
            if isinstance(fcs, np.ndarray):
                fcs = np.broadcast_to(fcs, bcast_shape + self.shape[self.stack_dim:])
            else:
                fcs = fcs.broadcast_to(bcast_shape + self.shape[self.stack_dim:])
            if isinstance(ocs, np.ndarray):
                ocs = np.broadcast_to(ocs, bcast_shape + other.shape[other.stack_dim:])
            else:
                ocs = ocs.broadcast_to(bcast_shape + other.shape[other.stack_dim:])
        else:
            fcs = self.coeffs
            ocs = other.coeffs
        return fcs, ocs
    @staticmethod
    def _enforce_ndarray_padding(fcs:np.ndarray, consistent_shape, pad_fcs, fc_padding):
        if fc_padding > 0:
            fcs = np.expand_dims(fcs, [-x for x in range(1, fc_padding + 1)])
        fcs = np.pad(fcs,
                     [[0, c - s] for c, s in zip(consistent_shape, pad_fcs)]
                     )
        return fcs
    @staticmethod
    def _enforce_sparse_padding(fcs:SparseArray, consistent_shape, pad_fcs, fc_padding):
        padding = [c - s for c, s in zip(consistent_shape, pad_fcs)]
        if fc_padding > 0:
            if np.sum(padding) == 0:
                fcs = fcs.expand_dims([-x for x in range(1, fc_padding + 1)])
            else:
                fcs = fcs.expand_and_pad(
                    [-x for x in range(1, fc_padding + 1)],
                    padding
                )
        elif np.sum(padding) > 0:
            fcs = fcs.pad_right(padding)
        return fcs
    @classmethod
    def _enforce_addition_padding(cls, fcs, consistent_shape, pad_fcs, fc_padding):
        if isinstance(fcs, np.ndarray):
            return cls._enforce_ndarray_padding(fcs, consistent_shape, pad_fcs, fc_padding)
        else:
            return cls._enforce_sparse_padding(fcs, consistent_shape, pad_fcs, fc_padding)
    def __add__(self, other)->'DensePolynomial':
        if isinstance(other, DensePolynomial):
            stack_dim = max([self.stack_dim, other.stack_dim])

            fcs, ocs = self._force_consistent_addition_shapes(other)
            if self.shape[self.stack_dim:] != other.shape[other.stack_dim:]:  # need to cast to consistent shape
                fc_padding = len(ocs.shape) - len(fcs.shape)
                oc_padding = -fc_padding
                pad_fcs = fcs.shape + tuple(1 for _ in range(fc_padding))
                pad_ocs = ocs.shape + tuple(1 for _ in range(oc_padding))
                consistent_shape = [
                    max(s1, s2)
                    for s1, s2 in zip(
                        pad_fcs,
                        pad_ocs
                    )
                ]

                fcs = self._enforce_addition_padding(fcs, consistent_shape, pad_fcs, fc_padding)
                ocs = self._enforce_addition_padding(ocs, consistent_shape, pad_ocs, oc_padding)

            try:
                fcs + ocs * (other.scaling / self.scaling)
            except:
                raise ValueError(fcs.shape, ocs.shape, (ocs*(other.scaling/self.scaling)).shape)
            return DensePolynomial(
                fcs + ocs*(other.scaling/self.scaling),
                shift=None,
                stack_dim=stack_dim,
                prefactor=self.scaling
            )
        else:
            if isinstance(other, (int, float, np.integer, np.floating)) and other == 0:
                return self

            if self.scaling != 1:
                other = other / self.scaling
            new = self._poly_coeffs.copy().flatten()
            new[0] += other
            new = new.reshape(self._poly_coeffs.shape)
            return DensePolynomial(
                new,
                shift=self._shift,
                prefactor=self.scaling,
                stack_dim=self.stack_dim
            )

    def shift(self, shift)->'DensePolynomial':
        if not isinstance(shift, (int, np.integer, float, np.floating)):
            shift = np.asanyarray(shift)
        return DensePolynomial(
            self._poly_coeffs,
            shift=(0 if self._shift is None else self._shift) + shift,
            prefactor=self._scaling,
            stack_dim=self.stack_dim
        )
    @classmethod
    def compute_shifted_coeffs(cls, poly_coeffs, shift, stack_dim=0):
        #TODO: make this consistent with non-zero stack_dim

        # if fourier_coeffs is None:
        #     raise ValueError("need fourier coeffs for shifted coeff calc")
        if isinstance(shift, (int, float, np.integer, np.floating)):
            shift = [shift]
        shift = np.asanyarray(shift)

        stack_shape = poly_coeffs.shape[:stack_dim]
        poly_shape = poly_coeffs.shape[stack_dim:]
        poly_dim = len(poly_shape)

        # handle broadcasting of shift
        if stack_dim > 0 and shift.ndim == 1 and shift.shape[0] == poly_dim:
            shift = np.expand_dims(shift, list(range(0, stack_dim)))#[0]*stack_dim)
        shift = np.broadcast_to(shift, stack_shape + (poly_dim,))

        # factorial outer product
        factorial_terms = np.array([math.factorial(x) for x in range(poly_shape[0])])
        for s in poly_shape[1:]:
            factorial_terms = np.expand_dims(factorial_terms, -1) * np.reshape(
                np.array([math.factorial(x) for x in range(s)]),
                [1]*factorial_terms.ndim + [s]
            )
        if stack_dim > 0:
            factorial_terms = np.broadcast_to(
                np.expand_dims(factorial_terms, list(range(0, stack_dim))),
                poly_coeffs.shape
            )

        if stack_dim > 0:
            shift = np.moveaxis(shift, -1, 0)
            shift_terms = np.power(
                shift[0][..., np.newaxis],
                np.expand_dims(np.arange(poly_shape[0]), list(range(0, stack_dim)))
            )
            for f,s in zip(shift[1:], poly_shape[1:]):
                shift_terms = np.expand_dims(shift_terms, -1) * np.reshape(
                    np.power(f[..., np.newaxis], np.expand_dims(np.arange(s), list(range(0, stack_dim)))),
                    stack_shape + (1,)*(shift_terms.ndim - stack_dim) + (s,)
                )
        else:
            shift_terms = np.power(shift[0], np.arange(poly_shape[0]))
            for f, s in zip(shift[1:], poly_shape[1:]):
                shift_terms = np.expand_dims(shift_terms, -1) * np.reshape(
                    np.power(f, np.arange(s)),
                    (1,) * shift_terms.ndim + (s,)
                )

        # build shifted polynomial coefficients
        rev_fac = factorial_terms
        for i in range(factorial_terms.ndim - stack_dim):
            rev_fac = np.flip(rev_fac, axis=stack_dim+i)
            shift_terms = np.flip(shift_terms, axis=stack_dim+i)

        poly_terms = poly_coeffs * factorial_terms
        shift_terms = shift_terms / rev_fac

        if not isinstance(poly_terms, np.ndarray):
            # poly_terms = poly_terms #type: SparseArray
            # raise Exception(
            #     len(poly_terms.block_vals) / np.prod(poly_terms.shape)
            # )
            poly_terms = poly_terms.asarray()

        if isinstance(poly_terms, np.ndarray):
            if stack_dim > 0:
                new = np.array([
                    scipy.signal.convolve(p, s)
                    for p,s in zip(
                        poly_terms.reshape((-1,)+poly_shape),
                        shift_terms.reshape((-1,)+poly_shape)
                    )
                ])
                new = new.reshape(stack_shape + new.shape[1:])
            else:
                new = scipy.signal.convolve(
                    poly_terms,
                    shift_terms
                )

            # stack_slice = (slice(None),)*stack_dim
            for i,s in enumerate(poly_shape):
                new = np.take(new, np.arange(-s, 0), axis=stack_dim+i)
            new = new / factorial_terms
        else:
            raise NotImplementedError(...)


        return new

    @classmethod
    def fill_tensors(self, tensors, idx, value, stack_dim, pcache, permute, rescale):
        # we compute the order of the index, compute the number of
        # permutations of the index, and then set all of the permutations
        # equal to the value divided by the number of perms
        order = sum(idx[stack_dim:])
        if order == 0:
            if stack_dim > 0:  # this corresponds to an element of only one of the polynomials
                tensors[0][idx[:stack_dim]] = value
            else:
                tensors[0] = value
        else:
            new_idx = sum(
                ([i] * n for i, n in enumerate(idx[stack_dim:])),
                []
            )  # we figure out which coords are affected and by how much
            if permute:
                idx_key = tuple(new_idx)
                if idx_key not in pcache:
                    final_perms = UniquePermutations(new_idx).permutations()
                    pcache[idx_key] = final_perms
                else:
                    final_perms = pcache[idx_key]
            else:
                final_perms = np.array([new_idx])
            # print(final_perms)
            if rescale:
                subvalue = value / len(final_perms)
            else:
                subvalue = value
            fp_inds = tuple(final_perms.T)
            if stack_dim > 0:  # this corresponds to an element of only one of the polynomials
                extra_idx = idx[:stack_dim]
                nels = len(fp_inds[0])
                bcast_idx = tuple(np.full(nels, i, dtype=int) for i in extra_idx)
                fp_inds = bcast_idx + fp_inds
            tensors[order][fp_inds] = subvalue
    @classmethod
    def extract_tensors(cls, coeffs, stack_dim=None, permute=True, rescale=True, cutoff=1e-15):
        if stack_dim is None:
            stack_dim = 0
            stack_shape = ()
        else:
            stack_shape = coeffs.shape[:stack_dim]

        if isinstance(coeffs, np.ndarray):
            nz_pos = np.where(np.abs(coeffs) > cutoff)
        else:
            nz_pos = coeffs.block_data[1]

        if len(nz_pos) == 0 or len(nz_pos[0]) == 0:
            return [0]
        max_order = np.max(np.sum(nz_pos[stack_dim:], axis=0, dtype=int))
        num_dims = len(coeffs.shape) - stack_dim
        tensors = [
            np.zeros(stack_shape + (num_dims,) * i)
            for i in range(1, max_order + 1)
        ]

        ref = [0] if stack_dim == 0 else [np.zeros(stack_shape)]
        tensors = ref + tensors  # include ref
        pcache = {}

        if isinstance(coeffs, np.ndarray):
            for idx in zip(*nz_pos): # TODO: speed this up...
                value = coeffs[idx]
                cls.fill_tensors(tensors, idx, value, stack_dim, pcache, permute, rescale)
        else:
            vals = coeffs.block_vals
            for vidx in zip(vals, *nz_pos):
                value, idx = vidx[0], vidx[1:]
                cls.fill_tensors(tensors, idx, value, stack_dim, pcache, permute, rescale)

        return tensors
    @classmethod
    def condense_tensors(cls, tensors, rescale=True, allow_sparse=True):
        # we'll be a bit slower here in constructing so we can make sure
        # we get the smallest tensor

        if not isinstance(tensors[0], (int, float, np.integer, np.floating)):
            tensors = [np.asanyarray(t) for t in tensors]
            stack_shape = np.asanyarray(tensors[0]).shape
        else:
            tensors = [tensors[0]] + [np.asanyarray(t) for t in tensors[1:]]
            stack_shape = ()
        stack_dim = len(stack_shape)

        # order = len(tensors)
        if len(tensors) == 1:
            raise NotImplementedError("constant polynomial support coming")

        ncoord = tensors[1].shape[-1]
        tensor_elems = {}
        shape = [1] * ncoord
        # we could be more efficient through where tricks if we wanted
        for o,t in enumerate(tensors):
            if o == 0:
                for idx in zip(*np.where(t != 0)):
                    tidx = idx + (0,)*ncoord
                    tensor_elems[tidx] = t[idx]
            else:
                utri_pos = tuple(np.array(list(it.combinations_with_replacement(range(ncoord), r=o))).T)
                if stack_dim > 0:
                    utri_pos = (slice(None),)*stack_dim + utri_pos
                vals = t[utri_pos]
                nz_pos = np.where(vals != 0)
                for p in zip(*nz_pos):
                    val = vals[p]
                    upos = [u[p[-1]] for u in utri_pos[stack_dim:]]
                    idx = tuple(np.bincount(upos, minlength=ncoord))
                    if rescale:
                        val = val * UniquePermutations.count_permutations(idx)
                    tidx = p[:-1] + idx
                    tensor_elems[tidx] = val
                    for n, i in enumerate(idx):
                        shape[n] = max(i + 1, shape[n])

        # raise Exception(shape)
        shp = stack_shape + tuple(shape)
        if allow_sparse:
            nv = len(tensor_elems)
            density = nv / np.prod(shp)
            allow_sparse = density < .5
        if allow_sparse:
            inds = tuple(np.array(list(tensor_elems.keys())).T)
            vals = np.array(list(tensor_elems.values()))
            condensed_tensor = SparseArray.from_data(
                (
                    vals,
                    inds
                ),
                shape=shp
            )
        else:
            condensed_tensor = np.zeros(shp)
            for idx,val in tensor_elems.items():
                condensed_tensor[idx] = val
        return condensed_tensor, stack_dim

    @property
    def coefficient_tensors(self):
        if self._coeff_tensors is None:
            self._coeff_tensors = self.extract_tensors(self.coeffs, stack_dim=self.stack_dim)
        return self._coeff_tensors
    @property
    def unscaled_coefficient_tensors(self):
        if self._uc_tensors is None:
            self._uc_tensors = self.extract_tensors(self.coeffs, rescale=False, stack_dim=self.stack_dim)
        return self._uc_tensors
    def transform(self, lin_transf):
        """
        Applies (for now) a linear transformation to the polynomial
        """
        #TODO: make this consistent with non-zero stack_dim
        if self.stack_dim != 0:
            raise NotImplementedError("need to make work with non-zero stack dims")

        tensors = self.coefficient_tensors
        transf_tensors = []
        for n,t in enumerate(tensors):
            for ax in range(n):
                t = np.tensordot(lin_transf, t, axes=[1, n-1])
            transf_tensors.append(t)
        new = self.condense_tensors(transf_tensors)
        return DensePolynomial(new)

    def outer(self, other):
        if self.stack_dim != 0 or other.stack_dim != 0:
            raise NotImplementedError("need to make work with non-zero stack dims")
        return DensePolynomial(
            np.multiply.outer(self.coeffs, other),
            prefactor=self.scaling*other.scaling
        )

    @classmethod
    def _coord_deriv(cls, coeffs, coord, stack_dim):
        coord = coord + stack_dim
        if coeffs.shape[coord] == 1:
            return 0
        n = coeffs.shape[coord]
        scaling = np.arange(1, n)
        scaling = np.expand_dims(scaling, list(range(coord)) + list(range(coord+1, coeffs.ndim)))
        new_coeffs = scaling * np.take(coeffs, np.arange(1, n), axis=coord)
        return new_coeffs
    def deriv(self, coord):
        if self._poly_coeffs.shape[self.stack_dim + coord] == 1:
            return 0
        new_coeffs = self._coord_deriv(self.coeffs, coord, self.stack_dim)
        return DensePolynomial(
            new_coeffs,
            prefactor=self.scaling,
            stack_dim=self.stack_dim
        )

    @classmethod
    def _apply_dense_grad(cls, coeffs, stack_dim ,shape, n, c):
        new_shape = shape[:stack_dim] + (n,) + c
        new_coeffs = np.zeros(new_shape)
        for i in range(n):
            sub_coeffs = cls._coord_deriv(coeffs, i, stack_dim)
            if not isinstance(sub_coeffs, np.ndarray):
                if sub_coeffs != 0: raise ValueError("...?")
                continue
            # construct slice spec for inserting into new coeffs
            insert_spec = (
                # [i, :, ..., :, :-1, :, ..., :]
                    (slice(None),) * stack_dim
                    + (i,)
                    + (slice(None),) * i
                    + (slice(None, c[i] - 1),)
                    + (slice(None),) * (n - (i + 1))
            )
            new_coeffs[insert_spec] = sub_coeffs

        return new_coeffs

    @classmethod
    def _take_sparse_position_derivs(cls, stack_dim, axis, nzvals, nzinds):
        axis = axis + stack_dim
        ax_inds = nzinds[axis]
        if np.sum(ax_inds) == 0: return 0 # quick fall through
        good_pos = np.where(ax_inds > 0)
        nzvals = nzvals[good_pos]
        ax_inds = ax_inds[good_pos]
        new_ax = ax_inds - 1
        nzvals = nzvals * ax_inds
        nzinds = tuple(
            nzinds[i][good_pos]
                if i != axis else
            new_ax
            for i in range(len(nzinds))
        )
        return nzvals, nzinds
    @classmethod
    def _apply_sparse_grad(cls, coeffs:SparseArray, stack_dim, shape, n, c):
        block_vals, block_inds = coeffs.block_data
        new_stuff = [
            cls._take_sparse_position_derivs(stack_dim, i, block_vals, block_inds)
            for i in range(n)
        ]
        if all(isinstance(x, int) and x==0 for x in new_stuff):
            return SparseArray.empty(shape[:stack_dim] + (n,))

        final_vals = np.concatenate([
            vals_inds[0] for vals_inds in new_stuff
            if not (isinstance(vals_inds, int) and vals_inds == 0)
        ])
        stack_inds = tuple(
            np.concatenate([
                vals_inds[1][i] for vals_inds in new_stuff
                if not (isinstance(vals_inds, int) and vals_inds == 0)
            ])
            for i in range(stack_dim)
        )
        new_inds = np.concatenate([
            np.full(len(vals_inds[0]), i, dtype='uint8')
            for i,vals_inds in enumerate(new_stuff)
            if not (isinstance(vals_inds, int) and vals_inds == 0)
        ])
        rem_inds = tuple(
            np.concatenate([
                vals_inds[1][i] for vals_inds in new_stuff
                if not (isinstance(vals_inds, int) and vals_inds == 0)
            ])
            for i in range(stack_dim, len(shape))
        )
        final_inds = stack_inds + (new_inds,) + rem_inds

        new_shape = shape[:stack_dim] + (n,) + c
        return SparseArray.from_data(
            (
                final_vals,
                final_inds
            ),
            shape=new_shape
        )

    @classmethod
    def _apply_grad(cls, coeffs, stack_dim):
        shape = coeffs.shape
        n = coeffs.ndim - stack_dim
        c = shape[stack_dim:]

        if isinstance(coeffs, SparseArray):
            new_coeffs = cls._apply_sparse_grad(coeffs, stack_dim, shape, n, c)
        else:
            new_coeffs = cls._apply_dense_grad(coeffs, stack_dim, shape, n, c)

        return new_coeffs

    def grad(self):

        return DensePolynomial(
            self._apply_grad(self.coeffs, self.stack_dim),
            prefactor=self._scaling,
            stack_dim=self.stack_dim+1
        )

    def clip(self, threshold=1e-15):
        clipped_inds = np.where(np.abs(self.coeffs * self.scaling) > threshold)
        if len(clipped_inds) == 0 or len(clipped_inds[0]) == 0:
            return 0
        new_shape = tuple(max(x) + 1 for x in clipped_inds)
        coeffs = np.zeros(new_shape)
        coeffs[clipped_inds] = self.coeffs[clipped_inds]
        return DensePolynomial(
            coeffs,
            prefactor=self._scaling,
            stack_dim=self.scaling
        )

    def make_sparse_backed(self, threshold=1e-15):
        base = SparseArray.from_data(self.clip(threshold).coeffs)
        return type(self)(
            base,
            prefactor=self._scaling,
            stack_dim=self.stack_dim
        )
    
class SparsePolynomial(AbstractPolynomial):
    """
    A semi-symbolic representation of a polynomial of tensor
    coefficients
    """
    def __init__(self, terms:dict, prefactor=1, ndim=None, canonicalize=True):
        self.terms = terms
        self.prefactor = prefactor
        self._ndim = ndim
        self._shape = None

    @property
    def scaling(self):
        return 1 if self.prefactor is None else self.prefactor
    @scaling.setter
    def scaling(self, s):
        self.prefactor = s

    def expand(self):
        if self.prefactor == 1:
            return self
        else:
            return type(self)({k:self.prefactor*v for k,v in self.terms.items()}, prefactor=1,
                              canonicalize=False)
    @classmethod
    def monomial(cls, idx, value=1):
        return cls({idx:value})
    def __repr__(self):
        return "{}({},{})".format(type(self).__name__, self.terms,self.prefactor)

    def __mul__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            if other == 1:
                return self
            elif other == 0:
                return SparsePolynomial({})
        if isinstance(other, SparsePolynomial):
            new_terms = {}
            for k,v in other.terms.items():
                for k2,v2 in self.terms.items():
                    t = tuple(sorted(k + k2))
                    new_terms[t] = new_terms.get(t, 0) + v * v2
                    if new_terms[t] == 0:
                        del new_terms[t]
            return type(self)(new_terms, prefactor=self.prefactor*other.prefactor)
        else:
            return type(self)(self.terms, prefactor=self.prefactor*other)
            # new_terms = {}
            # for k,v in self.terms.items():
            #     new_terms[k] = self.prefactor*other*v
            # return SparsePolynomial(new_terms)
    def __add__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            if other == 0:
                return self
            else:
                return self + type(self)({():other})
        else:
            self = self.expand()
            new_terms = {}
            if isinstance(other, SparsePolynomial):
                other = other.expand()
                for s in other.terms.keys() & self.terms.keys():
                    v = other.terms[s]
                    v2 = self.terms[s]
                    new = v + v2
                    if new != 0:
                        new_terms[s] = new
                for o in other.terms.keys() - self.terms.keys():
                    new_terms[o] = other.terms[o]
                for o in self.terms.keys() - other.terms.keys():
                    new_terms[o] = self.terms[o]
            else:
                for k2, v2 in self.terms.items():
                    new = v2 + other
                    if new != 0:
                        new_terms[k2] = new
            if len(new_terms) == 0:
                return 0
            return type(self)(new_terms, canonicalize=False)

    @classmethod
    def _to_tensor_idx(cls, term, ndim, tupleate=True):
        base = np.zeros(ndim, dtype=int)
        for idx, cnt in zip(*np.unique(term, return_counts=True)): #TODO: kinda dumb to do this 2x but w/e
            base[idx] = cnt
        if tupleate:
            return tuple(base)
        else:
            return base
        # idx_terms = {}
        # m = 0
        # for t in idx:
        #     m = max(m, t)
        #     idx_terms[t] = idx_terms.get(t, 0) + 1
        # return tuple(idx_terms[t] for t in range(m))
    @property
    def shape(self):
        if self._shape is None:
            max_dims = {}
            for t in self.terms.keys():
                for idx,cnt in zip(*np.unique(t, return_counts=True)): # can I assume sorted?
                    max_dims[idx] = max(max_dims.get(idx, 1), cnt + 1)
            if self._ndim is None:
                max_key = max(list(max_dims.keys()))
            else:
                max_key = self._ndim - 1
            self._shape = tuple(max_dims.get(k, 1) for k in range(max_key+1))
        return self._shape
    def as_dense(self)->DensePolynomial:
        new_coeffs = np.zeros(self.shape)
        ndim = new_coeffs.ndim
        for key, val in self.terms.items():
            idx = self._to_tensor_idx(key, ndim)
            new_coeffs[idx] = val
        return DensePolynomial(
            new_coeffs,
            prefactor=self.prefactor
        )

    def _get_sparse_shift_terms(self, term_vector, shift_vector) -> dict:
        """
        Provides the set of terms corresponding to shifting a term of the form
         `prod(x[i]**p[i])` (where `p` is the `term_vector`) by the corresponding `shift_vector`


        :param term_vector:
        :param shift_vector:
        :return:
        """

        max_dim = np.max(term_vector)
        if max_dim == 0:
            return {():1}
        bin_mat = Binomial(np.max(term_vector)+1)
        inds = np.moveaxis(np.indices(term_vector+1), 0, -1).reshape(-1, term_vector.shape[0])
        rp_binoms = bin_mat[term_vector[np.newaxis, :], inds]
        term_diffs = term_vector[np.newaxis, :] - inds
        rp_shifts = shift_vector[np.newaxis, :] ** term_diffs
        # rp_binoms[:, term_vector == 0] = 0 # shift shouldn't contribute here

        new_terms = {}
        full_terms = np.prod(rp_shifts*rp_binoms, axis=-1)
        for idx, t in zip(inds, full_terms):
            key = sum(( (i,)*n for i,n in enumerate(idx) ), ())
            new_terms[key] = t

        return new_terms
    def shift(self, shift)->'SparsePolynomial':
        shift = np.asanyarray(shift)

        new_terms = {}
        ndim = len(self.shape)
        for term,scaling in self.terms.items():
            idx = self._to_tensor_idx(term, ndim, tupleate=False)
            # print(ndim, idx, term)
            subterms = self._get_sparse_shift_terms(idx, shift)
            # print(scaling, subterms)
            for s in subterms.keys() & new_terms.keys():
                v = scaling*subterms[s]
                v2 = new_terms[s]
                new = v + v2
                if new != 0:
                    new_terms[s] = new
            for o in subterms.keys() - new_terms.keys():
                new_terms[o] = scaling*subterms[o]
            for o in new_terms.keys() - subterms.keys():
                new_terms[o] = new_terms[o]

        return type(self)(
            new_terms,
            prefactor=self.scaling,
            ndim=ndim
        )


class PureMonicPolynomial(SparsePolynomial):
    def __init__(self, terms: dict, prefactor=1, canonicalize=True):
        if canonicalize:
            new_terms = {}
            for k,v in terms.items():
                k = self.canonical_key(k)
                new_terms[k] = new_terms.get(k, 0) + v
            terms = new_terms
        super().__init__(terms, prefactor)

    @property
    def shape(self):
        raise ValueError("{} doesn't have a dense counterpart".format(type(self).__name__))
    def as_dense(self):
        raise ValueError("{} doesn't have a dense counterpart".format(type(self).__name__))
    def shift(self, shift) -> DensePolynomial:
        raise ValueError("{} doesn't have a dense counterpart".format(type(self).__name__))

    @classmethod
    def monomial(cls, idx, value=1):
        return cls({(idx,): value})

    @classmethod
    def key_hash(cls, monomial_tuple):
        # hashes tuples of indices to give a fast check that the tuples
        # are the same independent of ordering
        return sum(hash(x) for x in monomial_tuple)

    @classmethod
    @abc.abstractmethod
    def canonical_key(cls, monomial_tuple):
        raise NotImplementedError("{} is abstract".format(cls.__name__))

    def direct_multiproduct(self, other, key_value_generator):
        if not isinstance(other, PureMonicPolynomial):
            raise TypeError("doesn't make sense")

        new_terms = {}
        term_hashes = {}
        for k, v in self.terms.items():
            for k2, v2 in other.terms.items():
                for k3, v3 in key_value_generator(k, k2, v, v2):
                    # new_hash = self.key_hash(k3) #+ self.key_hash(k2)
                    # if new_hash in term_hashes:
                    #     t = term_hashes[new_hash]
                    #     if t != self.canonical_key(k3):
                    #         raise ValueError(k, k2, t, self.canonical_key(k3))
                    # else:
                    #     t = self.canonical_key(k3)  # construct new tuple and canonicalize
                    #     term_hashes[new_hash] = t
                    t = self.canonical_key(k3) # caching this led to collisions, need better tuple hashes
                    new_terms[t] = new_terms.get(t, 0) + v3  # this is the case where after mult. two terms merge
                    if new_terms[t] == 0:
                        del new_terms[t]
        if self.prefactor is None:
            pref = other.prefactor
        elif other.prefactor is None:
            pref = self.prefactor
        else:
            pref = self.prefactor * other.prefactor
        return type(self)(new_terms, prefactor=pref, canonicalize=False)
    def direct_product(self, other, key_func=None, mul=None):
        if isinstance(other, (int, float, np.integer, np.floating)):
            if other == 1:
                return self
            elif other == 0:
                return 0
                # return type(self)({})
        if isinstance(other, PureMonicPolynomial):
            if mul is None:
                mul = lambda a,b:a*b
            if key_func is None:
                key_func = lambda key1,key2:k+k2
            new_terms = {}
            term_hashes = {}
            for k,v in self.terms.items():
                for k2,v2 in other.terms.items():
                    new_hash = self.key_hash(k) + self.key_hash(k2)
                    if new_hash in term_hashes:
                        t = term_hashes[new_hash]
                    else:
                        t = self.canonical_key(key_func(k, k2)) # construct new tuple and canonicalize
                        term_hashes[new_hash] = t
                    new_terms[t] = new_terms.get(t, 0) + mul(v, v2) # this is the case where after mult. two terms merge
                    if new_terms[t] == 0:
                        del new_terms[t]
            if self.prefactor is None:
                pref = other.prefactor
            elif other.prefactor is None:
                pref = self.prefactor
            else:
                pref = self.prefactor*other.prefactor
            return type(self)(new_terms, prefactor=pref, canonicalize=False)
        else:
            return type(self)(self.terms, prefactor=self.prefactor*other if self.prefactor is not None else other, canonicalize=False)

    def rebuild(self, new_terms, prefactor=None, canonicalize=None):
        return type(self)(new_terms,
                          prefactor=self.prefactor if prefactor is None else prefactor,
                          canonicalize=False if canonicalize is None else canonicalize
                          )

    def filter(self, keys, mode='match'):
        if mode == 'match':
            return self._filter_exact(keys)
        elif mode == 'include':
            return self._filter_include(keys)
        elif mode == 'exclude':
            return self._filter_exclude(keys)
        else:
            raise ValueError("bad mode '{}'".format(mode))
    def _filter_exact(self, test_keys):
        subterms = {}
        test_terms = self.terms
        for t, v in test_terms.items():
            test_t = t
            for required_keys in test_keys:
                for k in required_keys:
                    if k in test_t:
                        i = test_t.index(k)
                        test_t = test_t[:i] + test_t[i + 1:]
                    else:
                        break
                else:
                    subterms[t] = v

        return self.rebuild(subterms)
    def _filter_include(self, test_keys):
        subterms = {}
        test_terms = self.terms
        test_keys = set(test_keys)
        for t, v in test_terms.items():
            for k in t:
                if k not in test_keys:
                    break
            else:
                subterms[t] = v

        return self.rebuild(subterms)
    def _filter_exclude(self, test_keys):
        subterms = {}
        test_terms = self.terms
        test_keys = set(test_keys)
        for t, v in test_terms.items():
            for k in test_keys:
                if k in t:
                    break
            else:
                subterms[t] = v

        return self.rebuild(subterms)

    def __mul__(self, other):
        return self.direct_product(other)
class TensorCoefficientPoly(PureMonicPolynomial):
    """
    Represents a polynomial constructed using tensor elements as monomials
    by tracking sets of indices
    """

    @classmethod
    def canonical_key(cls, monomial_tuple):
        # we need a way to sort indices, which we do by grouping by key length,
        # doing a standard sort for each length, and reconcatenating
        s_groups = {}
        for index_tuple in monomial_tuple:
            l = len(index_tuple)
            grp = s_groups.get(l, [])
            s_groups[l] = grp
            grp.append(index_tuple)
        t = tuple(
            grp
            for l in sorted(s_groups.keys())
            for grp in sorted(s_groups[l])
        )
        return t

# class TensorCoeffPoly(AbstractPolynomial):
#     """
#     A semi-symbolic representation of a polynomial of tensor
#     coefficients
#     """
#
#     def __init__(self, terms:dict, prefactor=1):
#         self.terms = terms
#         self.prefactor = prefactor
#     def expand(self):
#         if self.prefactor == 1:
#             return self
#         else:
#             return TensorCoeffPoly({k:self.prefactor*v for k,v in self.terms.items()}, prefactor=1)
#     @classmethod
#     def monomial(cls, idx, value=1):
#         return cls({(idx,):value})
#     def __repr__(self):
#         return "{}({},{})".format(type(self).__name__, self.terms,self.prefactor)



import subprocess, os, tempfile as tf

__all__ = [
    "ExternalProgramRunner"
]

class ExternalProgramRunner:

    default_opts = {}
    def __init__(self, binary,
                 parser=None, prefix=None, suffix=None, delete=True,
                 **runtime_opts
                 ):
        if os.path.isfile(binary):
            binary = os.path.abspath(binary)
        self.binary = binary
        self.parser = parser
        self.opts = dict(self.default_opts, prefix=prefix, suffix=suffix, delete=delete, **runtime_opts)

    class _write_dir:
        def __init__(self, dir=None, dir_prefix=None, dir_suffix=None, delete=True):
            self.dir = dir
            self._temp_dirs = []
            self.delete = delete
            self.opts = {'prefix':dir_prefix, 'suffix':dir_suffix}
        def __enter__(self):
            if self.dir is None:
                td = tf.TemporaryDirectory(**self.opts)
                self._temp_dirs.append(td)
                return td.__enter__()
            else:
                return self.dir
        def __exit__(self, exc_type, exc_val, exc_tb):
            if self.dir is None:
                td = self._temp_dirs.pop()
                td.__exit__(exc_type, exc_val, exc_tb)

    def prep_dir(self, dir):
        ...

    @classmethod
    def subprocess_run(cls, binary, input_file, **subprocess_opts):
        if os.path.isfile(binary):
            binary = os.path.abspath(binary)
        res = subprocess.run([binary, input_file], **subprocess_opts)
        return res

    text_file_extensions = ['.out', '.txt', '.log']
    @classmethod
    def _load_aux_file(cls, dir, file, delete):
        test = os.path.join(dir, file)
        if os.path.isfile(test):
            file = test
        if os.path.isfile(file):
            if delete:
                if os.path.splitext(file)[1] in cls.text_file_extensions:
                    mode = 'r'
                else:
                    mode = 'rb'
                with open(file, mode) as data:
                    # can't be sure if it's a binary file or not
                    return data.read()
            else:
                return file

    blacklist_files = [".DS_Store", ".git"]
    @classmethod
    def run_job(cls,
                binary, job,
                dir=None, dir_prefix=None, dir_suffix=None,
                mode='w',
                runner=None, prep_dir=None, prep_job=None, prep_results=None,
                return_auxiliary_files=True,
                prefix=None, suffix=None, delete=True,
                raise_errors=True,
                **subprocess_opts
                ):

        results = {}
        with cls._write_dir(dir=dir, dir_prefix=dir_prefix, dir_suffix=dir_suffix) as dir:
            if prep_dir is not None:
                prep_dir(dir)
            with tf.NamedTemporaryFile(dir=dir, mode=mode, prefix=prefix, suffix=suffix, delete=False) as inp:
                if not isinstance(job, str) and hasattr(job, 'format'):
                    job = job.format()
                if prep_job is not None:
                    job = prep_job(job)
                results['input_file'] = job
                inp.write(results['input_file'])
            try:
                if return_auxiliary_files is True:
                    existing_files = set(os.listdir(dir))
                else:
                    existing_files = set()
                if runner is None:
                    runner = cls.subprocess_run
                res = runner(binary, inp.name, **dict(capture_output=True, cwd=dir, **subprocess_opts))
                results['process_output'] = res
                if prep_results is not None:
                    results.update(prep_results(dir))
                if return_auxiliary_files is True:
                    for file in os.listdir(dir):
                        if file not in existing_files and file not in cls.blacklist_files:
                            results[file] = cls._load_aux_file(dir, file, delete)
                elif isinstance(return_auxiliary_files, dict):
                    for k,v in return_auxiliary_files.items():
                        data = cls._load_aux_file(dir, v.format(name=inp.name), delete)
                        if data is not None:
                            results[k] = data
                elif return_auxiliary_files:
                    if isinstance(return_auxiliary_files, str):
                        return_auxiliary_files = [return_auxiliary_files]
                    for v in return_auxiliary_files:
                        data = cls._load_aux_file(dir, v.format(name=inp.name), delete)
                        if data is not None:
                            results[v] = data
                err = res.stderr.decode().strip()
                if raise_errors and len(err) > 0:
                    raise IOError(err)

                if not delete:
                    return inp, results
                else:
                    return results

            finally:
                if delete:
                    try:
                        os.remove(inp.name)
                    except OSError:
                        pass
        # if dir is not None:

    def run(self, job,
            dir=None, dir_prefix=None, dir_suffix=None,
            mode=None,
            runner=None, prep_dir=None, prep_job=None, prep_results=None,
            return_auxiliary_files=None,
            prefix=None, suffix=None, delete=None,
            raise_errors=None,
            **job_opts):

        defaults = {
            k: v
            for k,v in dict(
                dir=dir, dir_prefix=dir_prefix, dir_suffix=dir_suffix,
                mode=mode,
                runner=runner, prep_dir=prep_dir, prep_job=prep_job, prep_results=prep_results,
                return_auxiliary_files=return_auxiliary_files,
                prefix=prefix, suffix=suffix, delete=delete,
                raise_errors=raise_errors
            ).items() if v is not None
        }
        return self.run_job(
            self.binary,
            job,
            **{
                **self.opts,
                'prep_dir':self.prep_dir,
                **defaults,
                **job_opts
            }
        )

        # with open(os.path.join(dir, 'atomicMasses.xml'), 'w+') as mass_file:
        #     mass_file.write(self.format_masses_file(self.atom_map) + "\n\n")

        # for key, ext in {"parallel": ".spectrum_parallel", "duschinsky": '.spectrum_dushinsky'}.items():
        #     test = inp.name + ext
        #     if os.path.isfile(test):
        #         with open(test) as strm:
        #             results[key] = strm.read()


"""
Provides support for chemical toolkits
"""
import io
import base64
from .Interface import *

__all__ = [
    "PILInterface",
    "OpenCVInterface"
]

class PILInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the OpenCV toolkit
    """
    name = 'PIL'
    module = 'PIL'
    @classmethod
    def from_file(cls, file, **opts):
        return cls.method("Image").open(file, **opts)

    @classmethod
    def from_url(cls, url):
        format, buf = cls.prep_url_buffer(url)
        if format is None:
            format = 'png'
        return cls.method('Image').open(buf)#, format=format)

    @classmethod
    def to_url(cls, image, format='png'):
        buf = io.BytesIO()
        image.save(buf, format=format)
        return base64.b64encode(buf.getvalue()).decode("ascii")

    @classmethod
    def prep_url_buffer(cls, img_data: str, format=None):
        if img_data.startswith('data:image/'):
            tag, img_data = img_data.split(";", 1)
            if format is None:
                format = tag.split("/")[1]

        b64_tag = 'base64,'
        if img_data.startswith(b64_tag):
            return format, io.BytesIO(base64.b64decode(img_data[len(b64_tag):]))
        else:
            return format, io.BytesIO(img_data.encode('ascii'))


class OpenCVInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the PIL toolkit
    """
    name = 'OpenCV'
    module = 'cv2'


from .Interface import *

__all__ = [
    "VPythonInterface",
    "VTKInterface",
    # "X3DInterface"
]

class VPythonInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the VPython visualization toolkit
    """
    name = 'VPython'
    module = 'vpython'

# class X3DInterface(ExternalProgramInterface):
#     """
#     A simple class to support operations that make use of the VPython visualization toolkit
#     """
#     name = 'X3D'
#     module = 'x3d'

class VTKInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the VPython visualization toolkit
    """
    name = 'VTK'
    module = 'vtk'

    @classmethod
    def graphics_object(cls, obj):
        return cls.method('vtk'+obj)()
    @classmethod
    def named_colors(cls):
        return cls.method('vtkNamedColors')()


import numpy as np
from .CommonData import DataHandler, DataRecord

__all__ = [ "ColorData" ]
__reload_hook__ = [".CommonData"]

class ColorDataHandler(DataHandler):
    def __init__(self):
        super().__init__("ColorData")#:, record_type=ColorDataRecord)

    def __getitem__(self, item):
        if isinstance(item, int):
            item = list(self.data['Palettes'].keys())[item]
            return super().__getitem__(('Palettes', item))
        elif isinstance(item, str) and item not in self.data:
            for k,v in self.data.items():
                if item in v:
                    return super().__getitem__((k, item))
            else:
                raise KeyError(f"couldn't find `ColorData` spec {item}")
        else:
            return super().__getitem__(item)

ColorData=ColorDataHandler()
ColorData.__doc__ = """An instance of `ColorDataHandler` that can be used for looking up data on color palettes"""
ColorData.__name__ = "ColorData"


"""
Defines classes for providing different approaches to fitting.
For the most part, the idea is to use `scipy.optimize` to do the actual fitting process,
but we layer on conveniences w.r.t. specification of bases and automation of the actual fitting process
"""
import abc

from .. import Devutils as dev

import numpy as np, scipy.optimize as opt, enum
from collections import OrderedDict as odict

__all__ = [
    "FittedModel"
]

class FittedModel:
    def __init__(self,
                 fit_basis,
                 expansion_coeffs=None,
                 basis_parameters=None,
                 **kwargs
                 ):
        self.fit_basis, self.basis_parameters = self.canonicalize_basis(fit_basis, basis_parameters)
        self.coeffs = expansion_coeffs
        self.opts = kwargs

    @classmethod
    def canonicalize_basis(cls, fit_basis, basis_parameters):
        if callable(fit_basis):
            fit_basis = [fit_basis]
        if basis_parameters is None or dev.is_dict_like(basis_parameters):
            basis_parameters = [basis_parameters] * len(fit_basis)

        return fit_basis, basis_parameters

    def __call__(self, pts, order=None, **opts):
        opts = dict(self.opts, **opts)
        return self.evaluate_kernel(
            self.fit_basis,
            self.basis_parameters,
            pts,
            coeffs=self.coeffs,
            order=order,
            **opts
        )

    @classmethod
    def evaluate_kernel(cls,
                        fit_basis,
                        basis_parameters,
                        pts,
                        coeffs=None,
                        order=None,
                        **opts
                        ):
        if order is None:
            kernel_expansions = [
                [
                    f(pts, **(params if params is not None else {}), **opts)
                    for f, params in zip(fit_basis, basis_parameters)
                ]
            ]
        else:
            kernel_expansions = [
                f(pts, order=order, **(params if params is not None else {}), **opts)
                for f, params in zip(fit_basis, basis_parameters)
            ]
            kernel_expansions = [
                [k[o] for k in kernel_expansions]
                for o in range(order + 1)
            ]

        if coeffs is not None:
            val_expansions = [np.dot(coeffs, k) for k in kernel_expansions]
        else:
            val_expansions = [sum(k) for k in kernel_expansions]

        if order is None:
            return val_expansions[0]
        else:
            return val_expansions

    @classmethod
    def _handle_nl_fit_params(cls,
                              params,
                              kernels,
                              param_names,
                              include_expansion_coefficients=True
                              ):
        if include_expansion_coefficients:
            n = len(kernels)
            coeffs, params = params[-n:], params[:-n]
        else:
            coeffs = None

        parameter_lists = []
        k = 0
        for pl in param_names:
            e = k+len(pl)
            parameter_lists.append(
                zip(pl, params[k:e])
            )
            k = e
        if k < len(params) - 1:
            raise ValueError(f"params of len {len(params)} don't distribute into names {param_names}")

        return coeffs, parameter_lists

    @classmethod
    def get_kernel_and_opts(cls, k):
        func, opts = k  # TODO: make this more flexible down the line
        if dev.is_dict_like(opts):
            names = list(opts.keys())
            vals = list(opts.values())
        else:
            names = list(opts)
            vals = [None] * len(names)
        vals = [
            v
                if v is not None else
            np.random.rand()
            for v in vals
        ]
        return func, names, vals
    @classmethod
    def parse_kernel_specs(cls, kernels):
        if dev.is_dict_like(kernels):
            kernels = [kernels]
        funcs = []
        param_names = []
        param_defaults = []
        for k in kernels:
            f, n, d = cls.get_kernel_and_opts(k)
            funcs.append(f)
            param_names.append(n)
            param_defaults.extend(d)

        return funcs, param_names, param_defaults

    @classmethod
    def nonlinear_fit(cls,
                      kernel_specs,
                      pts,
                      observations,
                      include_expansion_coefficients=True,
                      **fit_params
                      ):
        kernels, param_names, param_defaults = cls.parse_kernel_specs(kernel_specs)
        if include_expansion_coefficients:
            param_defaults = np.concatenate([np.array(param_defaults, dtype=float), np.ones(len(kernels), dtype=float)])
        def f(x, params):
            coeffs, basis_parameters = cls._handle_nl_fit_params(
                params,
                kernels,
                param_names,
                include_expansion_coefficients=include_expansion_coefficients
            )
            return cls.evaluate_kernel(
                kernels,
                basis_parameters,
                x,
                coeffs=coeffs
            )
        # def jac():

        opt_params, _ = opt.curve_fit(
            f,
            pts,
            observations,
            param_defaults,
            fit_params,
            full_output=False
        )

        coeffs, param_dicts = cls._handle_nl_fit_params(
            opt_params, kernels, param_names,
            include_expansion_coefficients=include_expansion_coefficients
        )

        return cls(
            kernels,
            expansion_coeffs=coeffs,
            basis_parameters=param_dicts
        )
        # return cls.from_fit(
        #     kernels,
        #     param_names,
        #     fit
        # )

    @classmethod
    def get_fit_methods(cls):
        return {
            'nonlinear_fit':cls.nonlinear_fit
        }

    _fit_dispatch = dev.uninitialized
    default_fit_method = 'nonlinear_fit'
    @classmethod
    def get_fit_dispatch(cls):
        cls._fit_dispatch = dev.handle_uninitialized(
            cls._fit_dispatch,
            dev.OptionsMethodDispatch,
            args=(cls.get_fit_methods,),
            kwargs=dict(
                default_method=cls.default_fit_method,
                # attributes_map=cls.get_evaluators_by_attributes()
            )
        )
        return cls._fit_dispatch
    @classmethod
    def fit(cls,
            kernels,
            pts,
            observations,
            method=None,
            **opts
            ):

        fit_method, method_opts = cls.get_fit_dispatch().resolve(method)
        return fit_method(
            kernels,
            pts,
            observations,
            **dict(method_opts, **opts)
        )







"""
Provides functionality for managing configurations stored in files
without having to write bespoke serialization/deserialization/application
code every time a new job type comes up
"""

import os, inspect
from .Serializers import *
from ..Devutils import OptionsSet

__all__ = [ "Config", "ParameterManager"]

class Config:
    """
    A configuration object which basically just supports
    a dictionary interface, but which also can automatically
    filter itself so that it only provides the keywords supported
    by a `from_config` method.
    """
    def __init__(self, config, serializer=None, extra_params=None):
        """Loads the config from a file
        :param config:
        :type config: str
        :param serializer:
        :type serializer: None | BaseSerializer
        """
        self.config = self.find_config(config)
        self._serializer = serializer
        self._conf = None
        self._loaded = False
        if extra_params is None:
            extra_params = {}
        self.extra = extra_params

    config_file_name = "config"
    config_file_extensions = [".json", ".yml", ".yaml", ".py"]
    @classmethod
    def find_config(self, config, name=None, extensions=None):
        """
        Finds configuration file (if config isn't a file)

        :param config:
        :type config:
        :return:
        :rtype:
        """
        if os.path.isdir(config):
            if name is None:
                name = self.config_file_name
            if extensions is None:
                extensions = self.config_file_extensions
            for ext in extensions:
                test = os.path.join(config, name + ext)
                if os.path.isfile(test):
                    return test
        elif os.path.isfile(config):
            return config

    _serializer_map = {
        ".py": ModuleSerializer,
        ".json": JSONSerializer,
        ".yaml": YAMLSerializer,
        ".yml": YAMLSerializer
    }
    @classmethod
    def get_serializer(self, file):
        _, ext = os.path.splitext(file)
        if ext not in self._serializer_map:
            raise ValueError("no known serializer for file extension {}".format(ext))
        mode = self._serializer_map[ext]
        return mode()

    @classmethod
    def new(cls, loc, init=None):
        config = os.path.join(loc,
                              cls.config_file_name + cls.config_file_extensions[0]
                              )
        serializer = cls.get_serializer(config)
        if init is None:
            init = {}
        with open(config, 'w') as stream:
            serializer.serialize(stream, init)
        return cls(config)

    def serialize(self, file, ops):
        if self._serializer is None:
            ser = self.get_serializer(file)
        else:
            ser = self._serializer
        with open(file, 'w') as stream:
            return ser.serialize(stream, ops)
    def deserialize(self, file):
        if self._serializer is None:
            ser = self.get_serializer(file)
        else:
            ser = self._serializer
        with open(file, 'r') as stream:
            return ser.deserialize(stream)

    def save(self):
        return self.serialize(self.config, self.opt_dict)
    def load(self):
        return self.deserialize(self.config)

    @property
    def name(self):
        try:
            res = self.get_key("name")
        except KeyError:
            res = os.path.basename(self.config)
        return res
    @property
    def opt_dict(self):
        self.load_opts()
        return dict(self._conf, **self.extra)

    def filter(self, keys, strict=True):
        """
        Returns a filtered option dictionary according to keys.
        Strict mode will raise an error if there is a key in the config that isn't
        in keys.

        :param keys:
        :type keys: Iterable[str] | function
        :param strict:
        :type strict: bool
        :return:
        :rtype:
        """
        try:
            keys = list(keys)
        except TypeError:
            # method we need to inspect to ask about kwargs
            import inspect
            args = inspect.signature(keys).parameters.values()
            keys = [arg.name for arg in args if arg.kind == arg.POSITIONAL_OR_KEYWORD]
        opts = self.opt_dict
        opt_keys = set(opts.keys())
        filt = {}
        for k in keys:
            if k in opts:
                filt[k] = opts[k]
                opt_keys.remove(k)
        if strict:
            if len(opt_keys) > 0:
                raise ValueError("{}: excess keys in config {}".format(
                    type(self).__name__,
                    opt_keys
                ))
        return filt

    def apply(self, func, strict=True):
        """
        Applies func to stored parameters

        :param func:
        :type func:
        :return:
        :rtype:
        """
        kw = self.filter(func, strict=strict)
        return func(**kw)

    def update(self, **kw):
        opts = self.opt_dict
        opts.update(**kw)
        self.save()

    def load_opts(self):
        if not self._loaded:
            cfg = self.config
            if cfg is None:
                raise ValueError("can't load config from None")
            self._conf = self.load()
            self._conf['config_location'] = os.path.dirname(cfg)

    def get_conf_attr(self, item):
        if not self._loaded:
            self.load_opts()
        if self._conf_type is dict:
            return self._conf_obj[item]
        else:
            return getattr(self._conf_obj, item)

    def __getattr__(self, item):
        return self.get_conf_attr(item)

class ParameterManager(OptionsSet):

    def serialize(self, file, mode = None):
        return ModuleSerializer().serialize(file, self.ops, mode = mode)

    @classmethod
    def deserialize(cls, file, mode=None, attribute=None):
        return ModuleSerializer().deserialize(file, mode=mode, attribute=attribute)


from .d3 import D3
from ..JHTML.WidgetTools import frozendict

__all__ = ["RendererD3", "FigureManagerD3", "FigureCanvasD3", "_BackendD3"]
__reload_hook__ = ["..JHTML", ".d3"]

import uuid, hashlib, itertools, io, base64, numpy as np
from PIL import Image

# from matplotlib import _api
# from matplotlib._pylab_helpers import Gcf
import matplotlib as mpl
from matplotlib import font_manager as fm
from matplotlib.backend_bases import (_Backend, FigureCanvasBase, FigureManagerBase, RendererBase)
from matplotlib.colors import rgb2hex
# from matplotlib.dates import UTC
from matplotlib.path import Path
from matplotlib import _path
from matplotlib.transforms import Affine2D, Affine2DBase

class RendererD3(RendererBase):
    """
    A modification of the base matplotlib SVG renderer to plug into the D3 library work we've done
    """

    _capstyle_d = {'projecting': 'square', 'butt': 'butt', 'round': 'round'}
    @staticmethod
    def _short_float_fmt(x):
        """
        Create a short string representation of a float, which is %f
        formatting with trailing zeros and the decimal point removed.
        """
        return f'{x:f}'.rstrip('0').rstrip('.')
    @staticmethod
    def _generate_transform(transform_list):
        parts = []
        for type, value in transform_list:
            if (type == 'scale' and (value == (1,) or value == (1, 1))
                    or type == 'translate' and value == (0, 0)
                    or type == 'rotate' and value == (0,)):
                continue
            if type == 'matrix' and isinstance(value, Affine2DBase):
                value = value.to_values()
            parts.append('{}({})'.format(type, ' '.join(RendererD3._short_float_fmt(x) for x in value)))
        return ' '.join(parts)

    def __init__(self, width, height, basename=None, image_dpi=72, *, metadata=None):
        self.width = width
        self.height = height
        self.image_dpi = image_dpi  # actual dpi at which we rasterize stuff

        if basename is None:
            basename = ""
        self.basename = basename

        self._tree = []
        self._elems = None
        self._toplevel = []
        self._defs = {'clips':{}, 'hatches':{}, 'glyphs':[], 'markers':[], 'paths':[]}
        self._groupd = {}
        self._image_counter = itertools.count()
        self._clipd = {}
        self._markers = {}
        self._path_collection_id = 0
        self._hatchd = {}
        # self._has_gouraud = False
        # self._n_gradients = 0

        super().__init__()
        self._glyph_map = dict()
        # str_height = _short_float_fmt(height)
        # str_width = _short_float_fmt(width)
        # svgwriter.write(svgProlog)
        # self._start_id = self.writer.start(
        #     'svg',
        #     width=f'{str_width}pt',
        #     height=f'{str_height}pt',
        #     viewBox=f'0 0 {str_width} {str_height}',
        #     xmlns="http://www.w3.org/2000/svg",
        #     version="1.1",
        #     attrib={'xmlns:xlink': "http://www.w3.org/1999/xlink"})
        # self._write_metadata(metadata)
        # self._write_default_style()

    def open_group(self, s, gid=None):
        # print(" "*len(self._tree), ">>", s)
        self._elems = []
        self._tree.append([s, gid, self._elems])
    def close_group(self, s):
        tag, gid, elems = self._tree.pop()
        # print(" "*len(self._tree), "<<", s)
        el = D3.Group(
            *elems,
            cls=tag.replace('.', '-'),
            id=gid
        )
        if len(self._tree) == 0:
            self._elems = None
            self._toplevel.append(el)
        else:
            self._elems = self._tree[-1][-1]
            self._elems.append(el)

    def write_defs(self):
        self._write_clips()
        self._write_hatches()
        # self.writer.close(self._start_id)
        # self.writer.flush()

    def _make_id(self, type, content):
        salt = mpl.rcParams['svg.hashsalt']
        if salt is None:
            salt = str(uuid.uuid4())
        m = hashlib.sha256()
        m.update(salt.encode('utf8'))
        m.update(str(content).encode('utf8'))
        return f'{type}{m.hexdigest()[:10]}'

    def _make_flip_transform(self, transform):
        return transform + Affine2D().scale(1, -1).translate(0, self.height)

    def _get_hatch(self, gc, rgbFace):
        """
        Create a new hatch pattern
        """
        if rgbFace is not None:
            rgbFace = tuple(rgbFace)
        edge = gc.get_hatch_color()
        if edge is not None:
            edge = tuple(edge)
        dictkey = (gc.get_hatch(), rgbFace, edge)
        oid = self._hatchd.get(dictkey)
        if oid is None:
            oid = self._make_id('h', dictkey)
            self._hatchd[dictkey] = ((gc.get_hatch_path(), rgbFace, edge), oid)
        else:
            _, oid = oid
        return oid

    def _write_hatches(self):
        if not len(self._hatchd):
            return
        hatch_data = []
        HATCH_SIZE = 72
        for hatch_key,(path, face, stroke), oid in self._hatchd.items():
            if hatch_key not in self._defs['hatches']:
                path_data = self._convert_path(
                    path,
                    Affine2D().scale(HATCH_SIZE).scale(1.0, -1.0).translate(0, HATCH_SIZE),
                    simplify=False
                )
                if face is None:
                    fill = 'none'
                else:
                    fill = rgb2hex(face)

                self._defs['hatches'][hatch_key] = D3.Pattern(
                        D3.Rect(x=0, y=0, width=HATCH_SIZE + 1, height=HATCH_SIZE + 1, fill=fill),
                        D3.Path(d=path_data, fill=rgb2hex(stroke), stroke=rgb2hex(stroke),
                                stroke_width=mpl.rcParams['hatch.linewidth'],
                                stroke_linecap='butt',
                                stroke_linejoin='miter'
                                ),
                        id=oid,
                        patternUnits="userSpaceOnUse",
                        x=0, y=0,
                        width=HATCH_SIZE,
                        height=HATCH_SIZE
                    )

    def _get_style_dict(self, gc, rgbFace):
        """Generate a style string from the GraphicsContext and rgbFace."""
        attrib = {}

        forced_alpha = gc.get_forced_alpha()

        if gc.get_hatch() is not None:
            attrib['fill'] = f"url(#{self._get_hatch(gc, rgbFace)})"
            if (rgbFace is not None and len(rgbFace) == 4 and rgbFace[3] != 1.0
                    and not forced_alpha):
                attrib['fill-opacity'] = rgbFace[3]
        else:
            if rgbFace is None:
                attrib['fill'] = 'none'
            else:
                if tuple(rgbFace[:3]) != (0, 0, 0):
                    attrib['fill'] = rgb2hex(rgbFace)
                if (len(rgbFace) == 4 and rgbFace[3] != 1.0
                        and not forced_alpha):
                    attrib['fill-opacity'] = rgbFace[3]

        if forced_alpha and gc.get_alpha() != 1.0:
            attrib['opacity'] = gc.get_alpha()

        offset, seq = gc.get_dashes()
        if seq is not None:
            attrib['stroke-dasharray'] = ','.join(val for val in seq)
            attrib['stroke-dashoffset'] = self._short_float_fmt(offset)

        linewidth = gc.get_linewidth()
        if linewidth:
            rgb = gc.get_rgb()
            attrib['stroke'] = rgb2hex(rgb)
            if not forced_alpha and rgb[3] != 1.0:
                attrib['stroke-opacity'] = rgb[3]
            if linewidth != 1.0:
                attrib['stroke-width'] = linewidth
            if gc.get_joinstyle() != 'round':
                attrib['stroke-linejoin'] = gc.get_joinstyle()
            if gc.get_capstyle() != 'butt':
                attrib['stroke-linecap'] = self._capstyle_d[gc.get_capstyle()]

        return attrib

    def _get_style(self, gc, rgbFace):
        return self._get_style_dict(gc, rgbFace)

    def _get_clip_attrs(self, gc):
        cliprect = gc.get_clip_rectangle()
        clippath, clippath_trans = gc.get_clip_path()
        if clippath is not None:
            clippath_trans = self._make_flip_transform(clippath_trans)
            dictkey = (id(clippath), str(clippath_trans))
        elif cliprect is not None:
            x, y, w, h = cliprect.bounds
            y = self.height-(y+h)
            dictkey = (x, y, w, h)
        else:
            return {}
        clip = self._clipd.get(dictkey)
        if clip is None:
            oid = self._make_id('p', dictkey)
            if clippath is not None:
                self._clipd[dictkey] = ((clippath, clippath_trans), oid)
            else:
                self._clipd[dictkey] = (dictkey, oid)
        else:
            clip, oid = clip
        return {'clip-path': f'url(#{oid})'}

    def _write_clips(self):
        if not len(self._clipd):
            return
        clip_paths = []
        for clip_key,(clip, oid) in self._clipd.items():
            if clip_key not in self._defs['clips']:
                if len(clip) == 2:
                    clippath, clippath_trans = clip
                    path_data = self._convert_path(clippath, clippath_trans, simplify=False)
                    clip_shape = D3.Path(d=path_data)
                else:
                    x, y, w, h = clip
                    clip_shape = D3.Rect(x=x, y=y, width=w, height=h)
                self._defs['clips'][clip_key] = D3.ClipPath(clip_shape, id=oid)

    def option_image_nocomposite(self):
        # docstring inherited
        return not mpl.rcParams['image.composite_image']

    def _convert_path(self, path, transform=None, clip=None, simplify=None, sketch=None):
        if clip:
            clip = (0.0, 0.0, self.width, self.height)
        else:
            clip = None
        return _path.convert_to_string(
            path, transform, clip, simplify, sketch, 6,
            [b'M', b'L', b'Q', b'C', b'z'], False).decode('ascii')

    def draw_path(self, gc, path, transform, rgbFace=None):
        # docstring inherited
        trans_and_flip = self._make_flip_transform(transform)
        clip = (rgbFace is None and gc.get_hatch_path() is None)
        simplify = path.should_simplify and clip
        path_data = self._convert_path(
            path, trans_and_flip, clip=clip, simplify=simplify,
            sketch=gc.get_sketch_params()
        )
        path = D3.Path(
            d=path_data,
            **self._get_clip_attrs(gc),
            style=self._get_style(gc, rgbFace)
        )
        if gc.get_url() is not None:
            path = D3.Link(
                path,
                **{'xlink:hred':gc.get_url()}
            )
        self._elems.append(path)

    def draw_markers(self, gc, marker_path, marker_trans, path, trans, rgbFace=None):
        if not len(path.vertices):
            return

        path_data = self._convert_path(
            marker_path,
            marker_trans + Affine2D().scale(1.0, -1.0),
            simplify=False)
        style = self._get_style_dict(gc, rgbFace)
        dictkey = (path_data, frozendict(style))
        oid = self._markers.get(dictkey)
        style = {k: v for k, v in style.items() if k.startswith('stroke')}

        if oid is None:
            oid = self._make_id('m', dictkey)
            self._defs['markers'].append(D3.Path(id=oid, d=path_data, style=style))
            self._markers[dictkey] = oid

        # writer.start('g', **self._get_clip_attrs(gc))
        trans_and_flip = self._make_flip_transform(trans)
        attrib = {'xlink:href': f'#{oid}'}
        clip = (0, 0, self.width*72, self.height*72)
        uses = []
        for vertices, code in path.iter_segments(trans_and_flip, clip=clip, simplify=False):
            if len(vertices):
                x, y = vertices[-2:]
                attrib['x'] = self._short_float_fmt(x)
                attrib['y'] = self._short_float_fmt(y)
                attrib['style'] = self._get_style(gc, rgbFace)
                uses.append(D3.Use(**attrib))
        self._elems.append(D3.Group(*uses, **self._get_clip_attrs(gc)))

    def draw_path_collection(self, gc, master_transform, paths, all_transforms,
                             offsets, offset_trans, facecolors, edgecolors,
                             linewidths, linestyles, antialiaseds, urls,
                             offset_position):
        # Is the optimization worth it? Rough calculation:
        # cost of emitting a path in-line is
        #    (len_path + 5) * uses_per_path
        # cost of definition+use is
        #    (len_path + 3) + 9 * uses_per_path
        len_path = len(paths[0].vertices) if len(paths) > 0 else 0
        uses_per_path = self._iter_collection_uses_per_path(
            paths, all_transforms, offsets, facecolors, edgecolors)
        should_do_optimization = len_path + 9 * uses_per_path + 3 < (len_path + 5) * uses_per_path
        if not should_do_optimization:
            return super().draw_path_collection(
                gc, master_transform, paths, all_transforms,
                offsets, offset_trans, facecolors, edgecolors,
                linewidths, linestyles, antialiaseds, urls,
                offset_position
            )

        path_codes = {}
        for i, (path, transform) in enumerate(
                self._iter_collection_raw_paths(master_transform, paths, all_transforms)
        ):
            transform = Affine2D(transform.get_matrix()).scale(1.0, -1.0)
            d = self._convert_path(path, transform, simplify=False)
            oid = 'C{:x}_{:x}_{}'.format(self._path_collection_id, i, self._make_id('', d))
            path_codes[oid] = D3.Path(id=oid, d=d)
        self._defs['paths'].extend(*path_codes.values())

        collections = []
        for xo, yo, path_id, gc0, rgbFace in self._iter_collection(
                gc, path_codes, offsets, offset_trans,
                facecolors, edgecolors, linewidths, linestyles,
                antialiaseds, urls, offset_position
        ):
            url = gc0.get_url()
            attrib = {
                'xlink:href': f'#{path_id}',
                'x': self._short_float_fmt(xo),
                'y': self._short_float_fmt(self.height - yo),
                'style': self._get_style(gc0, rgbFace)
                }
            elem = D3.Use(**attrib)
            clip_attrs = self._get_clip_attrs(gc0)
            if clip_attrs:
                elem = D3.Group(elem, **clip_attrs)
            if url is not None:
                elem = D3.Link(elem, **{'xlink:href': url})
            collections.append(elem)
        self._path_collection_id += 1
        self._elems.extend(collections)

    def option_scale_image(self):
        # docstring inherited
        return True

    def get_image_magnification(self):
        return self.image_dpi / 72.0

    def draw_image(self, gc, x, y, im, transform=None):

        h, w = im.shape[:2]

        if w == 0 or h == 0:
            return

        attrib = {}
        oid = gc.get_gid()
        # if mpl.rcParams['svg.image_inline']:
        buf = io.BytesIO()
        Image.fromarray(im).save(buf, format="png")
        oid = oid or self._make_id('image', buf.getvalue())
        attrib['xlink:href'] = (
            "data:image/png;base64,\n" +
            base64.b64encode(buf.getvalue()).decode('ascii')
        )
        # else:
        #     if self.basename is None:
        #         raise ValueError("Cannot save image data to filesystem when "
        #                          "writing SVG to an in-memory buffer")
        #     filename = f'{self.basename}.image{next(self._image_counter)}.png'
        #     # _log.info('Writing image file for inclusion: %s', filename)
        #     Image.fromarray(im).save(filename)
        #     oid = oid or 'Im_' + self._make_id('image', filename)
        #     attrib['xlink:href'] = filename
        attrib['id'] = oid

        if transform is None:
            w = 72.0 * w / self.image_dpi
            h = 72.0 * h / self.image_dpi

            elem = D3.Image(
                transform=self._generate_transform([('scale', (1, -1)), ('translate', (0, -h))]),
                x=x,
                y=-(self.height - y - h),
                width=w, height=h,
                **attrib
            )
        else:
            alpha = gc.get_alpha()
            if alpha != 1.0:
                attrib['opacity'] = alpha

            flipped = (
                Affine2D().scale(1.0 / w, 1.0 / h) +
                transform +
                Affine2D()
                .translate(x, y)
                .scale(1.0, -1.0)
                .translate(0.0, self.height)
            )

            attrib['transform'] = self._generate_transform([('matrix', flipped.frozen())])
            attrib['style'] = (
                'image-rendering:crisp-edges;'
                'image-rendering:pixelated')
            elem = D3.Image(width=w, height=h, **attrib)

        clip_attrs = self._get_clip_attrs(gc)
        if clip_attrs:
            # Can't apply clip-path directly to the image because the image has
            # a transformation, which would also be applied to the clip-path.
            elem = D3.Group(elem, **clip_attrs)

        url = gc.get_url()
        if url is not None:
            elem = D3.Link(elem, **{'xlink:href': url})

        self._elems.append(elem)

    def _update_glyph_map_defs(self, glyph_map_new):
        """
        Emit definitions for not-yet-defined glyphs, and record them as having
        been defined.
        """
        if glyph_map_new:
            new_glyphs = []
            for char_id, (vertices, codes) in glyph_map_new.items():
                char_id = self._adjust_char_id(char_id)
                # x64 to go back to FreeType's internal (integral) units.
                path_data = self._convert_path(
                    Path(vertices * 64, codes), simplify=False)
                new_glyphs.append(
                    D3.Path(
                        id=char_id,
                        d=path_data,
                        transform=self._generate_transform([('scale', (1 / 64,))])
                    )
                )
            self._defs['glyphs'].extend(new_glyphs)
            self._glyph_map.update(glyph_map_new)

    def _adjust_char_id(self, char_id):
        return char_id.replace("%20", "_")

    def _draw_text_as_path(self, gc, x, y, s, prop, angle, ismath, mtext=None):
        # docstring inherited

        # HTML.comment(s)

        glyph_map = self._glyph_map

        text2path = self._text2path
        color = rgb2hex(gc.get_rgb())
        fontsize = prop.get_size_in_points()

        style = {}
        if color != '#000000':
            style['fill'] = color
        alpha = gc.get_alpha() if gc.get_forced_alpha() else gc.get_rgb()[3]
        if alpha != 1:
            style['opacity'] = alpha
        font_scale = fontsize / text2path.FONT_SCALE
        text_group_attrib = {
            'style': style,
            'transform': self._generate_transform([
                ('translate', (x, y)),
                ('rotate', (-angle,)),
                ('scale', (font_scale, -font_scale))]),
        }

        paths = []
        if not ismath:
            font = text2path._get_font(prop)
            _glyphs = text2path.get_glyphs_with_font(
                font, s, glyph_map=glyph_map, return_new_glyphs_only=True)
            glyph_info, glyph_map_new, rects = _glyphs
            self._update_glyph_map_defs(glyph_map_new)

            for glyph_id, xposition, yposition, scale in glyph_info:
                attrib = {'xlink:href': f'#{glyph_id}'}
                if xposition != 0.0:
                    attrib['x'] = self._short_float_fmt(xposition)
                if yposition != 0.0:
                    attrib['y'] = self._short_float_fmt(yposition)
                paths.append(D3.Use(**attrib))

        else:
            if ismath == "TeX":
                _glyphs = text2path.get_glyphs_tex(
                    prop, s, glyph_map=glyph_map, return_new_glyphs_only=True)
            else:
                _glyphs = text2path.get_glyphs_mathtext(
                    prop, s, glyph_map=glyph_map, return_new_glyphs_only=True)
            glyph_info, glyph_map_new, rects = _glyphs
            self._update_glyph_map_defs(glyph_map_new)

            for char_id, xposition, yposition, scale in glyph_info:
                char_id = self._adjust_char_id(char_id)
                paths.append(
                    D3.Use(
                        transform=self._generate_transform([
                            ('translate', (xposition, yposition)),
                            ('scale', (scale,)),
                            ]),
                        **{'xlink:href': f'#{char_id}'}
                    )
                )

            for verts, codes in rects:
                path = Path(verts, codes)
                path_data = self._convert_path(path, simplify=False)
                paths.append(D3.Path(d=path_data))

        return D3.Group( # TODO: add comments
            *paths,
            **text_group_attrib
        )

    def _draw_text_as_text(self, gc, x, y, s, prop, angle, ismath, mtext=None):

        color = rgb2hex(gc.get_rgb())
        style = {}
        if color != '#000000':
            style['fill'] = color

        alpha = gc.get_alpha() if gc.get_forced_alpha() else gc.get_rgb()[3]
        if alpha != 1:
            style['opacity'] = alpha

        if not ismath:
            attrib = {}

            font_parts = []
            if prop.get_style() != 'normal':
                font_parts.append(prop.get_style())
            if prop.get_variant() != 'normal':
                font_parts.append(prop.get_variant())
            weight = fm.weight_dict[prop.get_weight()]
            if weight != 400:
                font_parts.append(f'{weight}')

            def _normalize_sans(name):
                return 'sans-serif' if name in ['sans', 'sans serif'] else name

            def _expand_family_entry(fn):
                fn = _normalize_sans(fn)
                # prepend generic font families with all configured font names
                if fn in fm.font_family_aliases:
                    # get all of the font names and fix spelling of sans-serif
                    # (we accept 3 ways CSS only supports 1)
                    for name in mpl.rcParams['font.' + fn]:
                        yield _normalize_sans(name)
                # whether a generic name or a family name, it must appear at
                # least once
                yield fn

            def _get_all_quoted_names(prop):
                # only quote specific names, not generic names
                return [name if name in fm.font_family_aliases else repr(name)
                        for entry in prop.get_family()
                        for name in _expand_family_entry(entry)]

            font_parts.extend([
                f'{self._short_float_fmt(prop.get_size())}px',
                # ensure expansion, quoting, and dedupe of font names
                ", ".join(dict.fromkeys(_get_all_quoted_names(prop)))
            ])
            style['font'] = ' '.join(font_parts)
            if prop.get_stretch() != 'normal':
                style['font-stretch'] = prop.get_stretch()
            attrib['style'] = style

            if mtext and (angle == 0 or mtext.get_rotation_mode() == "anchor"):
                # If text anchoring can be supported, get the original
                # coordinates and add alignment information.

                # Get anchor coordinates.
                transform = mtext.get_transform()
                ax, ay = transform.transform(mtext.get_unitless_position())
                ay = self.height - ay

                # Don't do vertical anchor alignment. Most applications do not
                # support 'alignment-baseline' yet. Apply the vertical layout
                # to the anchor point manually for now.
                angle_rad = np.deg2rad(angle)
                dir_vert = np.array([np.sin(angle_rad), np.cos(angle_rad)])
                v_offset = np.dot(dir_vert, [(x - ax), (y - ay)])
                ax = ax + v_offset * dir_vert[0]
                ay = ay + v_offset * dir_vert[1]

                ha_mpl_to_svg = {'left': 'start', 'right': 'end',
                                 'center': 'middle'}
                style['text-anchor'] = ha_mpl_to_svg[mtext.get_ha()]

                attrib['x'] = self._short_float_fmt(ax)
                attrib['y'] = self._short_float_fmt(ay)
                attrib['style'] = style
                attrib['transform'] = self._generate_transform([
                    ("rotate", (-angle, ax, ay))])

            else:
                attrib['transform'] = self._generate_transform([
                    ('translate', (x, y)),
                    ('rotate', (-angle,))])

            return D3.Text(s, **attrib)

        else:
            # writer.comment(s)

            width, height, descent, glyphs, rects = \
                self._text2path.mathtext_parser.parse(s, 72, prop)

            # Apply attributes to 'g', not 'text', because we likely have some
            # rectangles as well with the same style and transformation.
            group_attrs = dict(
                style=style,
                transform=self._generate_transform([
                    ('translate', (x, y)),
                    ('rotate', (-angle,))
                ]),
            )

            # Sort the characters by font, and output one tspan for each.
            spans = {}
            for font, fontsize, thetext, new_x, new_y in glyphs:
                entry = fm.ttfFontProperty(font)
                font_parts = []
                if entry.style != 'normal':
                    font_parts.append(entry.style)
                if entry.variant != 'normal':
                    font_parts.append(entry.variant)
                if entry.weight != 400:
                    font_parts.append(f'{entry.weight}')
                font_parts.extend([
                    f'{self._short_float_fmt(fontsize)}px',
                    f'{entry.name!r}',  # ensure quoting
                ])
                style = {'font': ' '.join(font_parts)}
                if entry.stretch != 'normal':
                    style['font-stretch'] = entry.stretch
                style = style
                if thetext == 32:
                    thetext = 0xa0  # non-breaking space
                spans.setdefault(style, []).append((new_x, -new_y, thetext))


            text_elems = []
            for style, chars in spans.items():
                chars.sort()

                if len({y for x, y, t in chars}) == 1:  # Are all y's the same?
                    ys = str(chars[0][1])
                else:
                    ys = ' '.join(str(c[1]) for c in chars)

                attrib = {
                    'style': style,
                    'x': ' '.join(self._short_float_fmt(c[0]) for c in chars),
                    'y': ys
                }

                text_elems.append(D3.TextSpan(
                    ''.join(chr(c[2]) for c in chars),
                    **attrib
                ))

            group_elems = [ D3.Text(*text_elems) ]
            for x, y, width, height in rects:
                group_elems.append(
                    D3.Rect(
                        x=x,
                        y=-y - 1,
                        width=width,
                        height=height
                    )
                )

            return D3.Group(
                *group_elems,
                **group_attrs
            )

    text_as_path = False
    def draw_text(self, gc, x, y, s, prop, angle, ismath=False, mtext=None):
        # docstring inherited

        if self.text_as_path:#mpl.rcParams['svg.fonttype'] == 'path':
            text_elem = self._draw_text_as_path(gc, x, y, s, prop, angle, ismath, mtext)
        else:
            text_elem = self._draw_text_as_text(gc, x, y, s, prop, angle, ismath, mtext)

        clip_attrs = self._get_clip_attrs(gc)
        if clip_attrs:
            # Cannot apply clip-path directly to the text, because
            # it has a transformation
            text_elem = D3.Group(text_elem, **clip_attrs)

        if gc.get_url() is not None:
            text_elem = D3.Link(text_elem, **{'xlink:href': gc.get_url()})

        self._elems.append(text_elem)

    def flipy(self):
        # docstring inherited
        return True

    def get_canvas_width_height(self):
        # docstring inherited
        return self.width, self.height

    def get_text_width_height_descent(self, s, prop, ismath):
        # docstring inherited
        return self._text2path.get_text_width_height_descent(s, prop, ismath)

    def get_toplevel(self):
        big_defs = []
        for def_list in self._defs.values():
            if isinstance(def_list, dict):
                big_defs.extend(def_list.values())
            else:
                big_defs.extend(def_list)
        elems = []
        if len(big_defs) > 0:
            elems.append(D3.Definitions(*big_defs))
        elems.extend(self._toplevel)

        return elems
    def insert_d3(self, root:'D3.Frame'):
        """
        width='%spt' % str_width,
        height='%spt' % str_height,
        viewBox='0 0 %s %s' % (str_width, str_height),
        xmlns="http://www.w3.org/2000/svg",
        version="1.1",
        attrib={'xmlns:xlink': "http://www.w3.org/1999/xlink"}
        :param root:
        :return:
        """
        elems = self.get_toplevel()
        root.svg.attrs = dict(
            width='{}pt'.format(self.width),
            height='{}pt'.format(self.height),
            viewBox='0 0 {w} {h}'.format(w=self.width, h=self.height),
            xmlns="http://www.w3.org/2000/svg",
            version="1.1"
            # attrib={'xmlns:xlink': "http://www.w3.org/1999/xlink"}
        )
        root.svg.elems = elems
        # for e in elems:
        #     root.append(e)

class FigureManagerD3(FigureManagerBase):
    """
    Manages a set of D3 canvases by providing a
    """

    def __init__(self, canvas:'FigureCanvasD3', num):
        # _log.debug("%s - __init__()", type(self))
        self.frame = D3.Frame()
        super().__init__(canvas, num)
        canvas.manager = self

    # @classmethod
    # def create_with_canvas(cls, canvas_class, figure, num):
    #     # docstring inherited
    #     frame = D3.Frame()
    #     return cls(canvas_class(figure=figure), frame, num)
    #     # return manager

    # @classmethod
    # def start_main_loop(cls):
    #     # no action needed, managed via Jupyter widgets
    #     ...

    def show(self):
        if not hasattr(self.canvas.figure, '_called_show') or self.canvas.figure._called_show:
            # self.frame.append(self.canvas.root)
            self.canvas.draw()
            self.frame.invalidate_cache()
            self.frame.display()

    # def destroy(self, *args):
    #     # no action needed, managed via Jupyter widgets
    #     ...

    def full_screen_toggle(self):
        raise NotImplementedError("can't make full screen...")
    def get_window_title(self):
        ...
        # # docstring inherited
        # return self.window.GetTitle()

    def set_window_title(self, title):
        ...
        # docstring inherited
        # self.window.SetTitle(title)

    def resize(self, width, height):
        dpi = self.canvas.figure.dpi
        self.frame.attr(height=height*dpi, width=width*dpi)

class FigureCanvasD3(FigureCanvasBase):
    """
    The canvas the figure renders into.  Calls the draw and print fig
    methods, creates the renderers, etc.

    Note: GUI templates will want to connect events for button presses,
    mouse movements and key presses to functions that call the base
    class methods button_press_event, button_release_event,
    motion_notify_event, key_press_event, and key_release_event.  See the
    implementations of the interactive backends for examples.

    Attributes
    ----------
    figure : `~matplotlib.figure.Figure`
        A high-level Figure instance
    """

    # The instantiated manager class.  For further customization,
    # ``FigureManager.create_with_canvas`` can also be overridden; see the
    # wx-based backends for an example.
    manager_class = FigureManagerD3

    def __init__(self, figure=None, manager=None):
        super().__init__(figure=figure)
        # self.root = D3.SVG()
        self.manager = manager

    def draw(self, clear=False):
        """
        Draw the figure using the renderer.

        It is important that this method actually walk the artist tree
        even if not output is produced because this will trigger
        deferred work (like computing limits auto-limits and tick
        values) that users may want access to before saving to disk.
        """
        wi, hi = self.figure.get_size_inches()
        dpi = self.figure.dpi
        renderer = RendererD3(wi*dpi, hi*dpi, image_dpi=self.figure.dpi)
        # print(self.figure.get_children()[1].get_children())
        self.figure.draw(renderer)
        # if clear:
        #     self.root.clear()
        renderer.insert_d3(self.manager.frame)

    @classmethod
    def render_objects(cls, figure, obj):
        wi, hi = figure.get_size_inches()
        dpi = figure.dpi
        renderer = RendererD3(wi * dpi, hi * dpi, image_dpi=figure.dpi)
        # print(self.figure.get_children()[1].get_children())
        if hasattr(obj, 'draw'):
            obj = [obj]
        for o in obj:
            o.draw(renderer)
        # if clear:
        #     self.root.clear()
        return renderer.get_toplevel()


########################################################################
#
# Now just provide the standard names that backend.__init__ is expecting
#
########################################################################

@_Backend.export
class _BackendD3(_Backend):
    FigureCanvas = FigureCanvasD3
    FigureManager = FigureManagerD3




import abc
import base64
import time
import urllib.parse
import collections
import weakref
from ..Scaffolding import ResourceManager
import hashlib
import zipfile
import os

__all__ = [
    "WebRequestHandler",
    "WebAPIConnection",
    "WebSubAPIConnection",
    "WebResourceManager",
    "GitHubReleaseManager",
    "ReleaseZIPManager"
]

class WebAPIError(IOError):
    ...

class WebRequestHandler:

    @classmethod
    def resolve_handler(cls, handler):
        if isinstance(handler, str):
            if handler == 'requests':
                import requests
                handler = cls.requests_request
            elif handler == 'urllib3':
                handler = cls.urllib3_request
            elif handler == 'default':
                handler = cls.default_request
            else:
                raise ValueError(f"don't know what to do with handler '{handler}'")
        return handler

    @classmethod
    def request(cls, method, url, json=None, handler=None, **params):
        response = None
        if json is not None and len(json) == 0:
            json = None
        if handler is not None:
            response = cls.resolve_handler(handler)(method, url, json=json, **params)
        else:
            try:
                response = cls.requests_request(method, url, json=json, **params)
            except ImportError:
                ...
            else:
                handler = 'requests'

            if response is None:
                try:
                    response = cls.urllib3_request(method, url, json=json, **params)
                except:
                    ...
                else:
                    handler = 'urllib3'

            if response is None:
                response = cls.default_request(method, url, json=json, **params)
                handler = 'default'

        return response, handler

    @classmethod
    def requests_request(cls, method, url, **params):
        import requests

        return requests.request(method, url, **params)

    @classmethod
    def urllib3_request(cls, method, url, **params):
        import urllib3

        return urllib3.request(method, url, **params)

    @classmethod
    def default_request(cls, method, url, data=None, headers=None, origin_req_host=None, unverifiable=False,
                        json=None, **params):
        import urllib.request, json as jsdump

        if json is not None:
            if data is None:
                data = jsdump.dumps(json).encode('utf-8')
            else:
                raise ValueError("can't get 'data' and 'json' arguments")


        return urllib.request.urlopen(
            urllib.request.Request(url,
                                   method=method,
                                   data=data,
                                   headers=headers,
                                   unverifiable=unverifiable,
                                   origin_req_host=origin_req_host
                                   ),
            **params
        )

    @classmethod
    def handle_response(cls, resp, headers):
        response, request_handler = resp
        if request_handler == 'requests':
            import requests
            response = response #type: requests.Request
            if response.status_code < 300:
                return response.json()
            else:
                raise WebAPIError(response.status_code, response.text, f"({response.url})")
        elif request_handler == 'urllib3':
            if response.status_code < 300:
                return response.json()
            else:
                raise WebAPIError(response.status_code, response.text, f"({response.url})")
        else:
            import json
            data = response.read()
            encoding = response.info().get_content_charset('utf-8')
            txt = data.decode(encoding)
            if response.status_code < 300:
                return json.loads(txt)
            else:
                raise WebAPIError(response.status_code, txt, f"({response.url})")

    @classmethod
    def read_response(cls, resp, decode=True):
        response, request_handler = resp
        if request_handler == 'requests':
            import requests
            response = response  # type: requests.Request
            if response.status_code < 300:
                if decode:
                    return response.text
                else:
                    return response.content
            else:
                raise WebAPIError(response.status_code, response.text, f"({response.url})")
        elif request_handler == 'urllib3':
            if response.status_code < 300:
                if decode:
                    return response.text
                else:
                    return response.content
            else:
                raise WebAPIError(response.status_code, response.text, f"({response.url})")
        else:
            data = response.read()
            if response.status_code < 300:
                if decode:
                    encoding = response.info().get_content_charset('utf-8')
                    return data.decode(encoding)
                else:
                    return data
            else:
                encoding = response.info().get_content_charset('utf-8')
                txt = data.decode(encoding)
                raise WebAPIError(response.status_code, txt, f"({response.url})")

class APIAuthentication(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def prep_request(self, url_params, **kwargs):
        ...

    @classmethod
    def resolve_auth(cls, auth_data):
        if hasattr(auth_data, 'prep_request'):
            return auth_data
        elif auth_data is None:
            return NoAuth()
        elif isinstance(auth_data, str):
            return BearerTokenAuth(auth_data)
        elif (
                not hasattr(auth_data, 'items')
                and len(auth_data) == 2
                and all(isinstance(x, str) for x in auth_data)
        ):
            return BasicAuth(*auth_data)
        else:
            return cls.dispatch_auth(auth_data)

    auth_types = {}
    @classmethod
    def get_auth_dispatch(cls):
        return dict({
            ('username', 'password'):BasicAuth,
            ('header', 'value'):HeaderValueAuth,
            ('token',):BearerTokenAuth,
        }, **cls.auth_types)
    @classmethod
    def dispatch_auth(cls, opts):
        for params, auth_type in sorted(
                cls.get_auth_dispatch().items(),
                key=lambda kt:-len(kt[0])
        ):
            if params is not None and all(p in opts for p in params):
                return auth_type(**opts)
        else:
            raise ValueError(f"can't dispatch auth type on {opts}")

class NoAuth(APIAuthentication):
    def prep_request(self, url_params, **kwargs):
        return url_params, kwargs

class HeaderValueAuth(APIAuthentication):
    def __init__(self, header, value):
        self.key = header
        self.val = value

    def prep_request(self, url_params, headers=None, **kwargs):
        if headers is None:
            headers = {}
        headers[self.key] = self.val
        return url_params, dict(kwargs, headers=headers)


class BasicAuth(APIAuthentication):
    """
    Does any site still use this???
    """
    def __init__(self, username, password):
        self.username = username
        self.password = password

    def prep_request(self, url_params, headers=None, **kwargs):
        if headers is None:
            headers = {}
        key = base64.urlsafe_b64encode(f'{self.username}:{self.password}'.encode()).decode()
        headers['Authentication'] = f'Basic {key}'
        return url_params, dict(kwargs, headers=headers)

class BearerTokenAuth(APIAuthentication):
    def __init__(self, token, encoded=True):
        if not encoded:
            base64.urlsafe_b64encode(token.encode()).decode()
        self.token = token

    def prep_request(self, url_params, headers=None, **kwargs):
        if headers is None:
            headers = {}
        headers['Authentication'] = f'Bearer {self.token}'
        return url_params, dict(kwargs, headers=headers)



class WebAPIConnection:
    """
    Base class for super simple web api interactions, use something better designed in general
    """

    def __init__(self, auth_info, history_length=None, log_requests=False, request_delay_time=None):
        self.auth = APIAuthentication.resolve_auth(auth_info)
        self.history = collections.deque(maxlen=history_length)
        self.subapi_instances = {}
        self.log_requests = log_requests
        self.delay = request_delay_time
        self._last_request = None

    default_content_type='application/json'
    default_return_type='application/json'
    def prep_headers(self, headers, content_type=None, return_type=None):
        if headers is None:
            headers = {}
        headers = headers.copy()
        if content_type is None:
            content_type = self.default_content_type
        if return_type is None:
            return_type = self.default_return_type
        if 'Content-Type' not in headers:
            headers['Content-Type'] = content_type
        if 'Accept' not in headers:
            headers['Accept'] = return_type
        return headers

    default_request_handler = WebRequestHandler
    def do_request(self,
                   method,
                   root,
                   *path,
                   query=None,
                   headers=None,
                   content_type=None,
                   return_type=None,
                   handler=None,
                   delay_time=None,
                   json=None,
                   data=None,
                   **urllib3_request_kwargs
                   ):
        if delay_time is None:
            delay_time = self.delay
        if delay_time is not None:
            if self._last_request is None:
                self._last_request = time.process_time() - delay_time
            proc_time = time.process_time()
            delay = proc_time - self._last_request
            if delay < self.delay:
                time.sleep(self.delay - delay)
            self._last_request = proc_time
        url_params = self.get_endpoint_params(root, path, query)
        url_params, urllib3_request_kwargs = self.auth.prep_request(url_params,
                                                                    headers=headers,
                                                                    json=json, data=data,
                                                                    **urllib3_request_kwargs)
        if 'headers' in urllib3_request_kwargs:
            headers = urllib3_request_kwargs['headers']
            del urllib3_request_kwargs['headers']
        if 'data' in urllib3_request_kwargs:
            data = urllib3_request_kwargs['data']
            del urllib3_request_kwargs['data']
        if 'json' in urllib3_request_kwargs:
            json = urllib3_request_kwargs['json']
            del urllib3_request_kwargs['json']

        headers = self.prep_headers(headers, content_type=content_type, return_type=return_type)
        url = urllib.parse.urlunsplit(url_params).rstrip("/")
        if handler is None:
            handler = self.default_request_handler
        if self.log_requests:
            print(method, url, json, headers)
        resp = handler.request(method, url,
                               headers=headers,
                               json=json, data=data,
                               **urllib3_request_kwargs)
        self.history.append(resp)
        return handler.handle_response(resp, headers)

    def get(self,
            root,
            *path,
            query=None,
            **urllib3_request_kwargs):
        return self.do_request('GET', root, *path, query=query, **urllib3_request_kwargs)

    def post(self,
             root,
             *path,
             query=None,
             **urllib3_request_kwargs):
        return self.do_request('POST', root, *path, query=query, **urllib3_request_kwargs)

    def delete(self,
               root,
               *path,
               query=None,
               **urllib3_request_kwargs):
        return self.do_request('DELETE', root, *path, query=query, **urllib3_request_kwargs)

    request_base = None
    def get_endpoint_params(self, root, path, query=None, base=None, fragment=None):
        if fragment is None:
            fragment = ''
        if query is None:
            query = ''
        elif not isinstance(query, str):
            query = urllib.parse.urlencode(query)
        if base is None:
            base = self.request_base
        if base is not None:
            root = urllib.parse.urljoin(base.rstrip('/')+"/", root)
        parse_root = urllib.parse.urlsplit(root)
        if isinstance(path, str):
            path = (path,)
        path = urllib.parse.urljoin(parse_root.path.rstrip('/')+"/",
                                    "/".join(urllib.parse.quote(s.strip("/")) for s in path)
                                    )
        return (
            parse_root.scheme,
            parse_root.netloc,
            path,
            query,
            fragment
        )

    def get_subapi(self, extension):
        if extension not in self.subapi_instances:
            self.subapi_instances[extension] = WebSubAPIConnection(extension, self)
        return self.subapi_instances[extension]

class WebSubAPIConnection(WebAPIConnection):
    def __init__(self, path_extension, root_api:WebAPIConnection):
        super().__init__(root_api.auth,
                         log_requests=root_api.log_requests,
                         request_delay_time=root_api.delay
                         )
        self.history = root_api.history
        self.request_base = urllib.parse.urljoin(root_api.request_base.rstrip('/')+"/", path_extension, allow_fragments=True)

class WebResourceManager(ResourceManager):
    default_resource_name = 'links'
    default_request_handler = WebRequestHandler
    def __init__(self, *, request_handler=None, **opts):
        if request_handler is None:
            request_handler = self.default_request_handler
        self.request_handler = request_handler
        super().__init__(**opts)
    def get_resource_filename(self, name):
        basename = urllib.parse.urlsplit(name).path.split('/')[-1]
        return '{}-{}'.format(basename, hashlib.md5(name.encode()).hexdigest())
    def download_link(self, link):
        return self.request_handler.read_response(
            self.request_handler.request("GET", link),
            decode=False
        )
    resource_function = download_link


class ReleaseZIPManager(WebResourceManager):
    default_resource_name = 'releases'
    location_env_var = 'GITHUB_RELEASE_DIR'
    use_temporary = True

    @classmethod
    def parse_semver(cls, version_string):
        return tuple(
            int(t)
            for t in version_string.rsplit('v', 1)[-1].split('.')
        )

    @classmethod
    def make_semver(cls, version):
        return 'v' + '.'.join(str(v) for v in version)

    @classmethod
    def parse_name_version(cls, filename):
        name, version = filename.rsplit('-', 1)[0].rsplit('v', 1)
        return name, cls.parse_semver(version)

    def list_resources(self):
        base_list = super().list_resources()
        resource_list = {}
        for filename, filepath in base_list.items():
            try:
                name, version = self.parse_name_version(filename)
            except ValueError:
                pass
            else:
                if name not in resource_list: resource_list[name] = {}
                resource_list[name][version] = filepath
        return resource_list
    def save_resource(self, loc, val):
        super().save_resource(loc, val)
        with zipfile.ZipFile(loc) as zip_extractor:
            dirs = zip_extractor.namelist()
            github_dirname = dirs[0]
            cwd = os.getcwd()
            try:
                os.chdir(self.location)
                zip_extractor.extractall(members=[d for d in dirs if d.startswith(github_dirname)])
                try:
                    os.remove(loc)
                except:
                    raise
                else:
                    os.rename(github_dirname, loc)
            finally:
                os.chdir(cwd)


class GitHubReleaseManager(WebAPIConnection):
    request_base = 'https://api.github.com/'
    resource_key = "zipball_url"
    release_manager_class = ReleaseZIPManager

    def __init__(self, token=None, request_delay_time=None, release_manager=None, **opts):
        if release_manager is None:
            release_manager = self.release_manager_class()
        self.release_manager = release_manager
        self.update_existing_releases()
        super().__init__({'header':'apikey', 'value':token}, request_delay_time=request_delay_time, **opts)


    blacklist_repos = [
        '.github'
    ]
    def list_repos(self, owner):
        return {
            r['name']:r
            for r in self.get('orgs', owner, 'repos')
            if r['name'] not in self.blacklist_repos
        }

    def list_releases(self, owner, repo):
        return {
            r['tag_name']:r
            for r in self.get('repos', owner, repo, 'releases')
        }
    def latest_release(self, owner, repo):
        return self.get('repos', owner, repo, 'releases', 'latest')

    # release_loc = None
    # def download_release(self, release_dict, where=None):
    #     with urllib.request.urlopen(release_dict['zipball_url']) as f:
    #         zip = f.read()

    release_cache = {}
    def update_existing_releases(self):
        release_list = self.release_manager.list_resources()
        for name, version_list in release_list.items():
            if name not in self.release_cache: self.release_cache[name] = {}
            self.release_cache[name].update(version_list)

    @classmethod
    def format_repo_key(cls, owner, name):
        return f"{owner}/{name}"
    def resolve_resource_url(self, v):
        return v[self.resource_key]
    def get_release_list(self, owner, name, update=None):
        key = self.format_repo_key(owner, name)
        if update or (update is None and key not in self.release_cache):
            self.release_cache[key] = {
                self.release_manager.parse_semver(k): self.resolve_resource_url(v)
                for k, v in self.list_releases(owner, name).items()
            }
        elif name not in self.release_cache:
            self.update_existing_releases()
        if name not in self.release_cache:
            raise ValueError(f"couldn't find any potential releases for {name}")
        return self.release_cache[name]


"""
Provides support for chemical toolkits
"""

from .Interface import *

__all__ = [
    "OpenBabelInterface",
    "PybelInterface",
    "RDKitInterface",
    "ASEInterface",
    "OpenChemistryInterface",
    "CCLibInterface"
]

class OpenBabelInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the OpenBabel toolkit (which is installed with anaconda)
    """
    name = 'OpenBabel'
    module = 'openbabel.openbabel'

class PybelInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the OpenBabel toolkit (which is installed with anaconda)
    """
    name = 'Pybel'
    module = 'openbabel.pybel'

class RDKitInterface(ExternalProgramInterface):
    name = 'RDKit'
    module = 'rdkit'

class ASEInterface(ExternalProgramInterface):
    name = 'ASE'
    module = 'ase'

    @classmethod
    def Atoms(cls,
              symbols: list[str] = None,
              positions: list[tuple[float, float, float]] = None,
              numbers: list[int] = None,
              masses: list[float] = None,
              charges: list[float] = None,
              **etc
              ):
        return cls.method('Atoms')(
            symbols=symbols,
            positions=positions,
            numbers=numbers,
            masses=masses,
            charges=charges,
            **etc
        )

class OpenChemistryInterface:
    name = 'OpenChemistry'
    module = 'openchemistry.io'

class PySCFInterface:
    name = 'PySCF'
    module = 'pyscf'

class CCLibInterface:
    name = 'CCLib'
    module = 'cclib'


import re

from ..Parsers import FileStreamReader, StringStreamReader
from ..Devutils import Logger, LogLevel, NullLogger, LoggingBlock
from ..Graphs import TreeWrapper

__all__ = [
    "Logger",
    "NullLogger",
    "LogLevel",
    "LogParser"
]


class LogParser(FileStreamReader):
    """
    A parser that will take a log file and stream it as a series of blocks
    """
    def __init__(self, file, block_settings=None, binary=False, block_level_padding=None, **kwargs):
        if block_settings is None:
            block_settings = LoggingBlock.block_settings
        self.block_settings = block_settings
        if block_level_padding is None:
            block_level_padding = LoggingBlock.block_level_padding
        self.block_level_padding = block_level_padding
        super().__init__(file, binary=binary, **kwargs)

    def get_block_settings(self, block_level):
        block_level_padding = self.block_level_padding
        if block_level >= len(self.block_settings):
            padding = block_level_padding * (block_level - len(self.block_settings) + 1)
            return {k: padding + v for k, v in self.block_settings[-1].items()}
        else:
            return self.block_settings[block_level]

    class LogBlockParser:
        """
        A little holder class that allows block data to be parsed on demand
        """
        def __init__(self, block_data, parent, block_depth):
            """
            :param block_data:
            :type block_data: str
            :param parent:
            :type parent: LogParser
            :param block_depth:
            :type block_depth: int
            """
            self.data = block_data
            self._lines = None
            self._tag = None
            self.parent = parent
            self.depth = block_depth

        @property
        def lines(self):
            if self._lines is None:
                self._tag, self._lines = self.parse_block_data()
            return self._lines
        @property
        def tag(self):
            if self._tag is None:
                self._tag, self._lines = self.parse_block_data()
            return self._tag

        def block_iterator(self, opener, closer,
                           preblock_handler=lambda c,w: w,
                           postblock_handler=lambda e:e,
                           start=0):

            where = self.data.find(opener, start)
            while where > -1:
                chunk = self.data[start:where]
                where = preblock_handler(chunk, where)
                end = self.data.find(closer, where)
                end = postblock_handler(end)
                subblock = self.data[where:end]
                start = end
                yield subblock, start
                where = self.data.find(opener, start)

        def line_iterator(self, pattern=""):
            og_settings = self.parent.get_block_settings(self.depth)
            prompt = og_settings['prompt'].format(meta="") + pattern
            raise NotImplementedError('Never finished the line iterator')

        def parse_prompt_blocks(self, chunk, prompt):
            splitsies = chunk.split("\n" + prompt)
            if splitsies[0] == "":
                splitsies = splitsies[1:]
            return splitsies

        def make_subblock(self, block):
            return type(self)(block, self.parent, self.depth+1)

        def parse_block_data(self):
            # find where subblocks are
            # parse around them
            og_settings = self.parent.get_block_settings(self.depth)
            prompt = og_settings['prompt'].split("{meta}", 1)[0]

            new_settings = self.parent.get_block_settings(self.depth+1)
            opener = "\n" + new_settings['opener'].split("{tag}", 1)[0]
            closer = "\n" + new_settings['closer'].split("{tag}", 1)[0]

            start = 0
            lines = []

            with StringStreamReader(self.data) as parser:
                header = parser.parse_key_block(None, {"tag":opener, "skip_tag":False})
                if header is not None:
                    lines.extend(self.parse_prompt_blocks(header, prompt))
                    block = parser.parse_key_block(None, {"tag":closer, "skip_tag":True})
                    if block is None:
                        raise ValueError("unclosed block found at position {} in stream '{}'".format(parser.tell(), parser.read()))
                    lines.append(self.make_subblock(block))
                    # print("??", parser.stream.read(1))
                    while header is not None:
                        header = parser.parse_key_block(None, {"tag":opener, "skip_tag":False})
                        curp = parser.tell()
                        if header is None:
                            break #
                        lines.extend(self.parse_prompt_blocks(header, prompt))
                        block = parser.parse_key_block(None, {"tag":closer, "skip_tag":True})
                        if block is None:
                            parser.seek(curp)
                            raise ValueError("unclosed block found at position {} in stream (from block '{}')".format(curp, parser.read(-1)))
                        lines.append(self.make_subblock(block))

                rest = parser.stream.read()
                lines.extend(self.parse_prompt_blocks(rest, prompt))

            tag_start = og_settings['opener'].split("{tag}", 1)[0]
            tag_end = og_settings['opener'].split("{tag}", 2)[-1]
            tag = lines[0]
            if tag_start != "":
                tag = tag.split(tag_start, 1)[-1]
            if tag_end != "":
                tag = tag.split(tag_end)[0]
            else:
                tag = tag.split("\n")[0]
            tag = tag.strip()

            block_end = "\n" + og_settings['closer'].split("{tag}", 1)[0]
            if isinstance(lines[-1], str):
                lines[-1] = lines[-1].split(block_end, 1)[0]

            return tag, lines[1:]

        def __repr__(self):
            return "{}('{}', #records={})".format(
                type(self).__name__,
                self.tag,
                len(self.lines)
            )

        def to_tree(self, tag_filter=None, depth=-1, combine_subtrees=True):
            tag, lines = self.tag, self.lines
            if depth < 0 or depth > 0:
                lines = [
                    b.to_tree(depth-1)
                        if not isinstance(b, str) else
                    b
                    for b in lines
                    if not (
                            not isinstance(b, str)
                            and self.parent.tag_match(b.tag, tag_filter)
                    )
                ]
                lines = self.parent.post_process_treelist(lines, combine_subtrees=combine_subtrees)

            if tag is not None:
                return {tag:lines}
            else:
                return lines

    def get_block(self, level=0, tag=None):
        """
        :param level:
        :type level:
        :param tag:
        :type tag:
        :return:
        :rtype:
        """

        block_settings = self.get_block_settings(level)
        if tag is None:
            opener = block_settings['opener'].split("{tag}", 1)[0]
        else:
            opener_split = block_settings['opener'].split("{tag}", 1)
            opener = opener_split[0]
            if len(opener_split) > 1:
                opener += " {} ".format(tag)

        if tag is None:
            closer = block_settings['closer'].split("{tag}", 1)[0]
        else:
            close_split = block_settings['closer'].split("{tag}", 1)
            closer = close_split[0]
            if len(close_split) > 1:
                closer += " {} ".format(tag)

        block_data = self.parse_key_block(opener, "\n"+closer, mode="Single", parser=lambda x:x) #type: str
        if block_data is None:
            raise ValueError("no more blocks")
        # I now need to process get_block further...
        block_data = opener + block_data.split("\n"+closer, 1)[0]

        return self.LogBlockParser(block_data, self, level)

    def get_line(self, level=0, tag=None):
        """
        :param level:
        :type level:
        :param tag:
        :type tag:
        :return:
        :rtype:
        """

        block_settings = self.get_block_settings(level)
        prompt = block_settings['prompt'].split("{meta}", 1)[0]
        if tag is not None:
            prompt += " {}".format(tag)

        # at some point I can try to refactor to keep the header info or whatever
        block_data = self.parse_key_block({'tag':"\n" + prompt, 'skip_tag':True}, {"tag":"\n", 'skip_tag': False}, mode="Single", parser=lambda x:x) #type: str
        if block_data is None:
            raise ValueError("no more lines")

        return block_data

    def get_blocks(self, tag=None, level=0):
        while True: # would be nice to have a smarter iteration protocol but ah well...
            try:
                next_block = self.get_block(level=level, tag=tag)
            except ValueError as e:
                args = e.args
                if len(args) == 1 and isinstance(args[0], str) and args[0] == "no more blocks":
                    return None
                raise
            else:
                if next_block is None:
                    return None
                yield next_block

    def get_lines(self, tag=None, level=0):
        while True: # would be nice to have a smarter iteration protocol but ah well...
            try:
                next_block = self.get_line(level=level, tag=tag)
            except ValueError as e:
                args = e.args
                if len(args) == 1 and isinstance(args[0], str) and args[0] == "no more lines":
                    return None
                raise
            else:
                if next_block is None:
                    return None
                yield next_block

    @classmethod
    def tag_match(cls, tag, tag_filter):
        if tag_filter is None or tag is None: return False
        elif isinstance(tag_filter, (str, re.Pattern)):
            return re.match(tag_filter, tag)
        elif callable(tag_filter):
            return tag_filter(tag)
        else:
            return tag in tag
    @classmethod
    def post_process_treelist(cls, res, combine_subtrees=True):
        if len(res) == 1:
            res = res[0]
        elif combine_subtrees and all(isinstance(r, dict) for r in res):
            nice_tree = {}
            key_conflicts = False
            for r in res:
                for k,v in r.items():
                    key_conflicts = k in nice_tree
                    if key_conflicts:
                        break
                    nice_tree[k] = v
                if key_conflicts:
                    break
            else:
                res = nice_tree
        return res
    def to_tree(self, tag_filter=None, depth=-1, combine_subtrees=True):
        res = []
        for b in self.get_blocks():
            if self.tag_match(b.tag, tag_filter): continue
            if depth < 0 or depth > 0:
                subtree = b.to_tree(tag_filter=tag_filter, depth=depth - 1, combine_subtrees=combine_subtrees)
            else:
                subtree = b
            res.append(subtree)

        return TreeWrapper(self.post_process_treelist(res, combine_subtrees=combine_subtrees))
